import os
import pathlib
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import pytest
import torch

import ludwig.error as ludwig_error
from ludwig.api import LudwigModel
from ludwig.constants import (
    ADAPTER,
    BACKEND,
    BASE_MODEL,
    BATCH_SIZE,
    EPOCHS,
    GENERATION,
    INPUT_FEATURES,
    MERGE_ADAPTER_INTO_BASE_MODEL,
    MODEL_LLM,
    MODEL_TYPE,
    OUTPUT_FEATURES,
    POSTPROCESSOR,
    PREPROCESSING,
    PRETRAINED_ADAPTER_WEIGHTS,
    PROGRESSBAR,
    PROMPT,
    QUANTIZATION,
    TRAINER,
    TYPE,
)
from ludwig.models.llm import LLM
from ludwig.schema.model_types.base import ModelConfig
from ludwig.utils.fs_utils import list_file_names_in_directory
from ludwig.utils.types import DataFrame
from tests.integration_tests.utils import category_feature, generate_data, text_feature

pytestmark = pytest.mark.llm


LOCAL_BACKEND = {"type": "local"}
TEST_MODEL_NAME = "hf-internal-testing/tiny-random-GPTJForCausalLM"
MAX_NEW_TOKENS_TEST_DEFAULT = 5

RAY_BACKEND = {
    "type": "ray",
    "processor": {
        "parallelism": 1,
    },
    "trainer": {
        "use_gpu": False,
        "num_workers": 2,
        "resources_per_worker": {
            "CPU": 1,
            "GPU": 0,
        },
    },
}


def get_num_non_empty_tokens(iterable):
    """Returns the number of non-empty tokens."""
    return len(list(filter(bool, iterable)))


@pytest.fixture(scope="module")
def local_backend():
    return LOCAL_BACKEND


@pytest.fixture(scope="module")
def ray_backend():
    return RAY_BACKEND


def get_dataset():
    data = [
        {"review": "I loved this movie!", "output": "positive"},
        {"review": "The food was okay, but the service was terrible.", "output": "negative"},
        {"review": "I can't believe how rude the staff was.", "output": "negative"},
        {"review": "This book was a real page-turner.", "output": "positive"},
        {"review": "The hotel room was dirty and smelled bad.", "output": "negative"},
        {"review": "I had a great experience at this restaurant.", "output": "positive"},
        {"review": "The concert was amazing!", "output": "positive"},
        {"review": "The traffic was terrible on my way to work this morning.", "output": "negative"},
        {"review": "The customer service was excellent.", "output": "positive"},
        {"review": "I was disappointed with the quality of the product.", "output": "negative"},
    ]
    df = pd.DataFrame(data)
    return df


def get_generation_config():
    return {
        "temperature": 0.1,
        "top_p": 0.75,
        "top_k": 40,
        "num_beams": 4,
        "max_new_tokens": MAX_NEW_TOKENS_TEST_DEFAULT,
    }


def convert_preds(preds: DataFrame):
    if isinstance(preds, pd.DataFrame):
        return preds.to_dict()
    return preds.compute().to_dict()


@pytest.mark.llm
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(LOCAL_BACKEND, id="local"),
        pytest.param(RAY_BACKEND, id="ray"),
    ],
)
def test_llm_text_to_text(tmpdir, backend, ray_cluster_4cpu):
    """Test that the LLM model can train and predict with text inputs and text outputs."""
    input_features = [
        {
            "name": "Question",
            "type": "text",
            "encoder": {"type": "passthrough"},
        }
    ]
    output_features = [text_feature(output_feature=True, name="Answer", decoder={"type": "text_extractor"})]

    csv_filename = os.path.join(tmpdir, "training.csv")
    dataset_filename = generate_data(input_features, output_features, csv_filename, num_examples=100)

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        GENERATION: get_generation_config(),
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        BACKEND: backend,
    }

    model = LudwigModel(config)
    model.train(dataset=dataset_filename, output_directory=str(tmpdir), skip_save_processed_input=True)

    preds, _ = model.predict(dataset=dataset_filename, output_directory=str(tmpdir), split="test")
    preds = convert_preds(preds)

    assert "Answer_predictions" in preds
    assert "Answer_probabilities" in preds
    assert "Answer_probability" in preds
    assert "Answer_response" in preds

    assert preds["Answer_predictions"]
    assert preds["Answer_probabilities"]
    assert preds["Answer_probability"]
    assert preds["Answer_response"]

    # Check that in-line generation parameters are used. Original prediction uses max_new_tokens = 5.
    assert get_num_non_empty_tokens(preds["Answer_predictions"][0]) <= MAX_NEW_TOKENS_TEST_DEFAULT
    original_max_new_tokens = model.model.generation.max_new_tokens

    # This prediction uses max_new_tokens = 2.
    preds, _ = model.predict(
        dataset=dataset_filename,
        output_directory=str(tmpdir),
        split="test",
        generation_config={"min_new_tokens": 2, "max_new_tokens": 3},
    )
    preds = convert_preds(preds)
    print(preds["Answer_predictions"][0])
    num_non_empty_tokens = get_num_non_empty_tokens(preds["Answer_predictions"][0])
    assert 2 <= num_non_empty_tokens <= 3

    # Check that the state of the model is unchanged.
    assert model.model.generation.max_new_tokens == original_max_new_tokens


@pytest.mark.llm
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(LOCAL_BACKEND, id="local"),
        pytest.param(RAY_BACKEND, id="ray"),
    ],
)
def test_llm_zero_shot_classification(tmpdir, backend, ray_cluster_4cpu):
    input_features = [
        {
            "name": "review",
            "type": "text",
        }
    ]
    output_features = [
        category_feature(
            name="output",
            preprocessing={
                "fallback_label": "neutral",
            },
            # How can we avoid using r here for regex, since it is technically an implementation detail?
            decoder={
                "type": "category_extractor",
                "match": {
                    "positive": {"type": "contains", "value": "positive"},
                    "neutral": {"type": "regex", "value": r"\bneutral\b"},
                    "negative": {"type": "contains", "value": "negative"},
                },
            },
        )
    ]

    df = get_dataset()

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        GENERATION: get_generation_config(),
        PROMPT: {"task": "This is a review of a restaurant. Classify the sentiment."},
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        BACKEND: backend,
    }

    model = LudwigModel(config)
    model.train(dataset=df, output_directory=str(tmpdir), skip_save_processed_input=True)

    prediction_df = pd.DataFrame(
        [
            {"review": "The food was amazing!", "output": "positive"},
            {"review": "The service was terrible.", "output": "negative"},
            {"review": "The food was okay.", "output": "neutral"},
        ]
    )

    preds, _ = model.predict(dataset=prediction_df, output_directory=str(tmpdir))
    preds = convert_preds(preds)

    assert preds


@pytest.mark.llm
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(LOCAL_BACKEND, id="local"),
        pytest.param(RAY_BACKEND, id="ray"),
    ],
)
def test_llm_few_shot_classification(tmpdir, backend, csv_filename, ray_cluster_4cpu):
    input_features = [
        text_feature(
            output_feature=False,
            name="body",
            encoder={"type": "passthrough"},  # need to use the default encoder for LLMTextInputFeatureConfig
        )
    ]
    output_features = [
        category_feature(
            output_feature=True,
            name="output",
            preprocessing={
                "fallback_label": "3",
            },
            decoder={
                "type": "category_extractor",
                "match": {
                    "1": {"type": "contains", "value": "1"},
                    "2": {"type": "contains", "value": "2"},
                    "3": {"type": "contains", "value": "3"},
                    "4": {"type": "contains", "value": "4"},
                    "5": {"type": "contains", "value": "5"},
                },
            },
        )
    ]

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        GENERATION: get_generation_config(),
        PROMPT: {
            "retrieval": {"type": "random", "k": 3},
            "task": (
                "Given the sample input, complete this sentence by replacing XXXX: The review rating is XXXX. "
                "Choose one value in this list: [1, 2, 3, 4, 5]."
            ),
        },
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        PREPROCESSING: {
            "split": {TYPE: "fixed"},
        },
        BACKEND: {**backend, "cache_dir": str(tmpdir)},
    }

    dataset_path = generate_data(
        input_features,
        output_features,
        filename=csv_filename,
        num_examples=25,
        nan_percent=0.1,
        with_split=True,
    )
    df = pd.read_csv(dataset_path)
    df["output"] = np.random.choice([1, 2, 3, 4, 5], size=len(df)).astype(str)  # ensure labels match the feature config
    df.to_csv(dataset_path, index=False)

    model = LudwigModel(config)
    model.train(dataset=dataset_path, output_directory=str(tmpdir), skip_save_processed_input=True)

    # TODO: fix LLM model loading
    # model = LudwigModel.load(os.path.join(results.output_directory, "model"), backend=backend)
    preds, _ = model.predict(dataset=dataset_path)
    preds = convert_preds(preds)

    assert preds


def _prepare_finetuning_test(
    csv_filename: str, finetune_strategy: str, backend: Dict, adapter_args: Dict
) -> Tuple[Dict, str]:
    input_features = [text_feature(name="input", encoder={"type": "passthrough"})]
    output_features = [text_feature(name="output")]

    train_df = generate_data(input_features, output_features, filename=csv_filename, num_examples=25)
    prediction_df = pd.DataFrame(
        [
            {"input": "The food was amazing!", "output": "positive"},
            {"input": "The service was terrible.", "output": "negative"},
            {"input": "The food was okay.", "output": "neutral"},
        ]
    )

    model_name = TEST_MODEL_NAME
    if finetune_strategy == "adalora":
        # Adalora isn't supported for GPT-J model types, so use tiny bart
        model_name = "hf-internal-testing/tiny-random-BartModel"
    elif finetune_strategy == "adaption_prompt":
        # At the time of writing this test, Adaption Prompt fine-tuning is only supported for Llama models
        model_name = "yujiepan/llama-2-tiny-random"

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: model_name,
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        BACKEND: backend,
    }

    if finetune_strategy is not None:
        config[ADAPTER] = {
            TYPE: finetune_strategy,
            **adapter_args,
        }

    return train_df, prediction_df, config


def _finetune_strategy_requires_cuda(finetune_strategy_name: str, quantization_args: Union[dict, None]) -> bool:
    """This method returns whether a given finetine_strategy requires CUDA.

    For all finetune strategies, except "qlora", the decision is based just on the name of the finetine_strategy; in the
    case of qlora, if the quantization dictionary is non-empty (i.e., contains quantization specifications), then the
    original finetine_strategy name of "lora" is interpreted as "qlora" and used in the lookup, based on the list of
    finetine strategies requiring CUDA.
    """
    cuda_only_finetune_strategy_names: List[str] = [
        "prompt_tuning",
        "prefix_tuning",
        "p_tuning",
        "qlora",
    ]

    if finetune_strategy_name == "lora" and quantization_args:
        finetune_strategy_name = "qlora"

    return finetune_strategy_name in cuda_only_finetune_strategy_names


def _verify_lm_lora_finetuning_layers(
    attention_layer: torch.nn.Module,
    merge_adapter_into_base_model: bool,
    model_weights_directory: str,
    expected_lora_in_features: int,
    expected_lora_out_features: int,
    expected_file_names: List[str],
) -> None:
    """This method verifies that LoRA finetuning layers have correct types and shapes, depending on whether the
    optional "model.merge_and_unload()" method (based on the "merge_adapter_into_base_model" directive) was
    executed.

    If merge_adapter_into_base_model is True, then both LoRA projection layers, V and Q, in the attention layer must
    contain square weight matrices (with the dimensions expected_lora_in_features by expected_lora_in_features).
    However, if merge_adapter_into_base_model is False, then the LoRA part of the attention layer must include Lora_A
    and Lora_B children layers for each of V and Q projections, such that the product of V and Q matrices is a square
    matrix (with the dimensions expected_lora_in_features by expected_lora_in_features) for both V and Q projections.
    """
    file_names: List[str] = list_file_names_in_directory(directory_name=model_weights_directory)
    assert set(file_names) == set(expected_file_names)
    assert isinstance(attention_layer.v_proj, torch.nn.Linear)
    assert isinstance(attention_layer.q_proj, torch.nn.Linear)
    if merge_adapter_into_base_model:
        assert (attention_layer.v_proj.in_features, attention_layer.v_proj.out_features) == (
            expected_lora_in_features,
            expected_lora_out_features,
        )
        assert (attention_layer.q_proj.in_features, attention_layer.q_proj.out_features) == (
            expected_lora_in_features,
            expected_lora_out_features,
        )
        assert not list(attention_layer.v_proj.children())
        assert not list(attention_layer.q_proj.children())
    else:
        v_proj_named_children: dict[str, torch.nn.Modeule] = dict(attention_layer.v_proj.named_children())
        assert isinstance(v_proj_named_children["lora_A"]["default"], torch.nn.Linear)
        assert (
            v_proj_named_children["lora_A"]["default"].in_features,
            v_proj_named_children["lora_A"]["default"].out_features,
        ) == (expected_lora_in_features, expected_lora_out_features)
        assert isinstance(v_proj_named_children["lora_B"]["default"], torch.nn.Linear)
        assert (
            v_proj_named_children["lora_B"]["default"].in_features,
            v_proj_named_children["lora_B"]["default"].out_features,
        ) == (expected_lora_out_features, expected_lora_in_features)
        q_proj_named_children: dict[str, torch.nn.Modeule] = dict(attention_layer.q_proj.named_children())
        assert isinstance(q_proj_named_children["lora_A"]["default"], torch.nn.Linear)
        assert (
            q_proj_named_children["lora_A"]["default"].in_features,
            q_proj_named_children["lora_A"]["default"].out_features,
        ) == (expected_lora_in_features, expected_lora_out_features)
        assert isinstance(q_proj_named_children["lora_B"]["default"], torch.nn.Linear)
        assert (
            q_proj_named_children["lora_B"]["default"].in_features,
            q_proj_named_children["lora_B"]["default"].out_features,
        ) == (expected_lora_out_features, expected_lora_in_features)


# TODO(arnav): p-tuning and prefix tuning have errors when enabled that seem to stem from DDP:
#
# prefix tuning:
# Sizes of tensors must match except in dimension 1. Expected size 320 but got size 32 for tensor number 1 in the list.
#
# p-tuning:
# 'PromptEncoder' object has no attribute 'mlp_head'
@pytest.mark.llm
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(LOCAL_BACKEND, id="local"),
        # TODO(Arnav): Re-enable once we can run tests on GPUs
        # This is because fine-tuning requires Ray with the deepspeed strategy, and deepspeed
        # only works with GPUs
        # pytest.param(RAY_BACKEND, id="ray"),
    ],
)
@pytest.mark.parametrize(
    "finetune_strategy,adapter_args",
    [
        pytest.param(
            None,
            {},
            id="full",
        ),
        pytest.param(
            "lora",
            {},
            id="lora-defaults",
        ),
        pytest.param(
            "lora",
            {"r": 4, "dropout": 0.1},
            id="lora-modified-defaults",
        ),
        pytest.param(
            "lora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: True, PROGRESSBAR: True}},
            id="lora_merged",
        ),
        pytest.param(
            "lora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: False}},
            id="lora_not_merged",
        ),
        pytest.param(
            "adalora",
            {},
            id="adalora-defaults",
        ),
        pytest.param(
            "adalora",
            {"init_r": 8, "beta1": 0.8},
            id="adalora-modified-defaults",
        ),
        pytest.param(
            "adalora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: True, PROGRESSBAR: True}},
            id="adalora_merged",
        ),
        pytest.param(
            "adalora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: False}},
            id="adalora_not_merged",
        ),
        pytest.param(
            "adaption_prompt",
            {},
            id="adaption_prompt-defaults",
        ),
        pytest.param(
            "adaption_prompt",
            {"adapter_len": 6, "adapter_layers": 1},
            id="adaption_prompt-modified-defaults",
        ),
        # pytest.param(
        #     "prompt_tuning",
        #     {
        #         "num_virtual_tokens": 8,
        #         "prompt_tuning_init": "RANDOM",
        #     },
        #     id="prompt_tuning_init_random",
        # ),
        # pytest.param(
        #     "prompt_tuning",
        #     {
        #         "num_virtual_tokens": 8,
        #         "prompt_tuning_init": "TEXT",
        #         "prompt_tuning_init_text": "Classify if the review is positive, negative, or neutral: ",
        #     },
        #     id="prompt_tuning_init_text",
        # ),
        # pytest.param(
        #     "prefix_tuning",
        #     {
        #         "num_virtual_tokens": 8,
        #     },
        #     id="prefix_tuning",
        # ),
        # pytest.param(
        #     "p_tuning",
        #     {"num_virtual_tokens": 8, "encoder_reparameterization_type": "MLP"},
        #     id="p_tuning_mlp_reparameterization",
        # ),
        # pytest.param(
        #     "p_tuning",
        #     {"num_virtual_tokens": 8, "encoder_reparameterization_type": "LSTM"},
        #     id="p_tuning_lstm_reparameterization",
        # ),
    ],
)
def test_llm_finetuning_strategies(tmpdir, csv_filename, backend, finetune_strategy, adapter_args):
    train_df, prediction_df, config = _prepare_finetuning_test(csv_filename, finetune_strategy, backend, adapter_args)

    output_directory: str = str(tmpdir)
    model_directory: str = pathlib.Path(output_directory) / "api_experiment_run" / "model"

    model = LudwigModel(config)
    model.train(dataset=train_df, output_directory=output_directory, skip_save_processed_input=False)

    # Make sure we can load the saved model and then use it for predictions
    model = LudwigModel.load(str(model_directory), backend=backend)

    base_model = LLM(ModelConfig.from_dict(config))
    assert not _compare_models(base_model, model.model)  # noqa F821

    preds, _ = model.predict(dataset=prediction_df, output_directory=output_directory)
    preds = convert_preds(preds)

    assert preds


@pytest.mark.llm
@pytest.mark.parametrize(
    "finetune_strategy,adapter_args,quantization",
    [
        pytest.param(
            "lora",
            {},
            {"bits": 4},
            id="qlora-4bit",
        ),
        pytest.param(
            "lora",
            {},
            {"bits": 8},
            id="qlora-8bit",
        ),
    ],
)
def test_llm_finetuning_strategies_quantized(tmpdir, csv_filename, finetune_strategy, adapter_args, quantization):
    if (
        _finetune_strategy_requires_cuda(finetune_strategy_name=finetune_strategy, quantization_args=quantization)
        and not (torch.cuda.is_available() and torch.cuda.device_count()) > 0
    ):
        pytest.skip("Skip: quantization requires GPU and none are available.")

    backend = LOCAL_BACKEND

    train_df, prediction_df, config = _prepare_finetuning_test(csv_filename, finetune_strategy, backend, adapter_args)
    config["backend"] = backend
    config[QUANTIZATION] = quantization

    model = LudwigModel(config)
    model.train(dataset=train_df, output_directory=str(tmpdir), skip_save_processed_input=False)

    # Make sure we can load the saved model and then use it for predictions
    model = LudwigModel.load(os.path.join(str(tmpdir), "api_experiment_run", "model"))

    base_model = LLM(ModelConfig.from_dict(config))
    assert not _compare_models(base_model, model.model)  # noqa F821

    preds, _ = model.predict(dataset=prediction_df, output_directory=str(tmpdir))
    preds = convert_preds(preds)

    assert preds


@pytest.mark.llm
@pytest.mark.skipif(torch.cuda.device_count() == 0, reason="test requires at least 1 gpu")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="test requires gpu support")
@pytest.mark.parametrize(
    "finetune_strategy,adapter_args,quantization,error_raised",
    [
        pytest.param(
            "lora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: False}},
            {"bits": 4},
            (
                ImportError,
                "Using `load_in_8bit=True` requires Accelerate: `pip install accelerate` and the latest version of bitsandbytes `pip install -i https://test.pypi.org/simple/ bitsandbytes` or pip install bitsandbytes` ",  # noqa E501
            ),
            id="qlora-4bit-not-merged",
        ),
        pytest.param(
            "lora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: True, PROGRESSBAR: True}},
            {"bits": 8},
            (
                ImportError,
                "Using `load_in_8bit=True` requires Accelerate: `pip install accelerate` and the latest version of bitsandbytes `pip install -i https://test.pypi.org/simple/ bitsandbytes` or pip install bitsandbytes` ",  # noqa E501
            ),
            id="qlora-8bit-merged",
        ),
        pytest.param(
            "lora",
            {POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: False}},
            {"bits": 8},
            (
                ImportError,
                "Using `load_in_8bit=True` requires Accelerate: `pip install accelerate` and the latest version of bitsandbytes `pip install -i https://test.pypi.org/simple/ bitsandbytes` or pip install bitsandbytes` ",  # noqa E501
            ),
            id="qlora-8bit-not-merged",
        ),
    ],
)
def test_llm_lora_finetuning_merge_and_unload_quantized_accelerate_required(
    csv_filename, finetune_strategy, adapter_args, quantization, error_raised
):
    input_features: List[dict] = [text_feature(name="input", encoder={"type": "passthrough"})]
    output_features: List[dict] = [text_feature(name="output")]

    config: dict = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        ADAPTER: {
            TYPE: finetune_strategy,
            **adapter_args,
        },
        QUANTIZATION: quantization,
    }

    model = LudwigModel(config)

    error_class: type  # noqa [F842]  # incorrect flagging of "local variable is annotated but never used
    error_message: str  # noqa [F842]  # incorrect flagging of "local variable is annotated but never used
    error_class, error_message = error_raised
    with pytest.raises(error_class) as excinfo:
        train_df = generate_data(input_features, output_features, filename=csv_filename, num_examples=3)
        model.train(dataset=train_df)

    assert str(excinfo.value) == error_message


@pytest.mark.llm
def test_llm_lora_finetuning_merge_and_unload_4_bit_quantization_not_supported(local_backend: dict):
    input_features: List[dict] = [text_feature(name="input", encoder={"type": "passthrough"})]
    output_features: List[dict] = [text_feature(name="output")]
    finetune_strategy: str = "lora"

    config: dict = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        ADAPTER: {
            TYPE: finetune_strategy,
            POSTPROCESSOR: {MERGE_ADAPTER_INTO_BASE_MODEL: True, PROGRESSBAR: True},
        },
        QUANTIZATION: {"bits": 4},
        BACKEND: local_backend,
    }

    expected_error_class: type = ludwig_error.ConfigValidationError
    expected_error_message: str = """This operation will entail merging LoRA layers on a 4-bit quantized model.  \
Calling "save_pretrained()" on that model is currently unsupported.  If you want to merge the LoRA adapter weights \
into the base model, you need to use 8-bit quantization or do non-quantized based training by removing the \
quantization section from your Ludwig configuration."""
    with pytest.raises(expected_error_class) as excinfo:
        _ = LudwigModel(config)

    assert str(excinfo.value) == expected_error_message


@pytest.mark.llm
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param(LOCAL_BACKEND, id="local"),
        # TODO: Re-enable once we can run tests on GPUs
        # This is because fine-tuning requires Ray with the deepspeed strategy, and deepspeed
        # only works with GPUs
        # pytest.param(RAY_BACKEND, id="ray"),
    ],
)
@pytest.mark.parametrize(
    "merge_adapter_into_base_model,expected_lora_in_features,expected_lora_out_features,expected_file_names",
    [
        pytest.param(
            False,
            32,
            8,
            [
                "README.md",
                "adapter_config.json",
                "adapter_model.bin",
            ],
            id="lora_not_merged",
        ),
        pytest.param(
            True,
            32,
            32,
            [
                "README.md",
                "adapter_config.json",
                "adapter_model.bin",
                "added_tokens.json",
                "config.json",
                "generation_config.json",
                "merges.txt",
                "pytorch_model.bin",
                "special_tokens_map.json",
                "tokenizer.json",
                "tokenizer_config.json",
                "vocab.json",
            ],
            id="lora_merged",
        ),
    ],
)
def test_llm_lora_finetuning_merge_and_unload(
    tmpdir,
    csv_filename,
    backend,
    merge_adapter_into_base_model,
    expected_lora_in_features,
    expected_lora_out_features,
    expected_file_names,
):
    finetune_strategy: str = "lora"
    adapter_args: dict = {
        POSTPROCESSOR: {
            MERGE_ADAPTER_INTO_BASE_MODEL: merge_adapter_into_base_model,
        },
    }
    train_df, prediction_df, config = _prepare_finetuning_test(
        csv_filename=csv_filename, finetune_strategy=finetune_strategy, backend=backend, adapter_args=adapter_args
    )

    output_directory: str = str(tmpdir)
    model_directory: str = pathlib.Path(output_directory) / "api_experiment_run" / "model"
    model_weights_directory: str = pathlib.Path(output_directory) / "api_experiment_run" / "model" / "model_weights"

    model = LudwigModel(config)
    model.train(dataset=train_df, output_directory=output_directory, skip_save_processed_input=False)
    _verify_lm_lora_finetuning_layers(
        attention_layer=model.model.model.base_model.model.transformer.h[1].attn,
        merge_adapter_into_base_model=merge_adapter_into_base_model,
        model_weights_directory=model_weights_directory,
        expected_lora_in_features=expected_lora_in_features,
        expected_lora_out_features=expected_lora_out_features,
        expected_file_names=expected_file_names,
    )

    # Make sure we can load the saved model and verify that the LoRA layers have expected shapes.
    model = LudwigModel.load(str(model_directory), backend=backend)
    _verify_lm_lora_finetuning_layers(
        attention_layer=model.model.model.base_model.model.transformer.h[1].attn,
        merge_adapter_into_base_model=merge_adapter_into_base_model,
        model_weights_directory=model_weights_directory,
        expected_lora_in_features=expected_lora_in_features,
        expected_lora_out_features=expected_lora_out_features,
        expected_file_names=expected_file_names,
    )


@pytest.mark.llm
@pytest.mark.parametrize("use_adapter", [True, False], ids=["with_adapter", "without_adapter"])
def test_llm_training_with_gradient_checkpointing(tmpdir, csv_filename, use_adapter):
    input_features = [text_feature(name="input", encoder={"type": "passthrough"})]
    output_features = [text_feature(name="output")]

    df = generate_data(input_features, output_features, filename=csv_filename, num_examples=25)

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: "hf-internal-testing/tiny-random-BartModel",
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 1,
            "enable_gradient_checkpointing": True,
        },
    }

    if use_adapter:
        config[ADAPTER] = {TYPE: "lora"}

    model = LudwigModel(config)
    assert model.config_obj.trainer.enable_gradient_checkpointing

    model.train(dataset=df, output_directory=str(tmpdir), skip_save_processed_input=False)


@pytest.mark.llm
def test_lora_wrap_on_init():
    from peft import PeftModel
    from transformers import PreTrainedModel

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
    }
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)
    assert isinstance(model.model, PreTrainedModel)
    assert not isinstance(model.model, PeftModel)

    # Now add adapter
    config[ADAPTER] = {
        TYPE: "lora",
    }
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)
    # We need to explicitly make this call since we now load the adapter
    # in the trainer as opposed to the point of LLM model initialization.
    model.prepare_for_training()
    assert not isinstance(model.model, PreTrainedModel)
    assert isinstance(model.model, PeftModel)


def test_llama_rope_scaling():
    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: "HuggingFaceH4/tiny-random-LlamaForCausalLM",
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        "model_parameters": {
            "rope_scaling": {
                "type": "dynamic",
                "factor": 2.0,
            }
        },
    }
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)

    assert model.model.config.rope_scaling
    assert model.model.config.rope_scaling["type"] == "dynamic"
    assert model.model.config.rope_scaling["factor"] == 2.0


def test_default_max_sequence_length():
    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: TEST_MODEL_NAME,
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
        TRAINER: {
            TYPE: "finetune",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        ADAPTER: {TYPE: "lora", PRETRAINED_ADAPTER_WEIGHTS: "Infernaught/test_adapter_weights"},
        BACKEND: {TYPE: "local"},
    }
    config_obj = ModelConfig.from_dict(config)
    assert config_obj.input_features[0].preprocessing.max_sequence_length is None
    assert config_obj.output_features[0].preprocessing.max_sequence_length is None


@pytest.mark.llm
@pytest.mark.parametrize("adapter", ["lora", "adalora", "adaption_prompt"])
def test_load_pretrained_adapter_weights(adapter):
    from peft import PeftModel
    from transformers import PreTrainedModel

    if adapter == "lora":
        weights = "Infernaught/test_adapter_weights"
        base_model = TEST_MODEL_NAME
    elif adapter == "adalora":
        weights = "Infernaught/test_adalora_weights"
        base_model = "HuggingFaceH4/tiny-random-LlamaForCausalLM"
    elif adapter == "adaption_prompt":
        weights = "Infernaught/test_ap_weights"
        base_model = "HuggingFaceH4/tiny-random-LlamaForCausalLM"
    else:
        raise ()

    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: base_model,
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
        TRAINER: {
            TYPE: "none",
            BATCH_SIZE: 8,
            EPOCHS: 2,
        },
        ADAPTER: {TYPE: adapter, PRETRAINED_ADAPTER_WEIGHTS: weights},
        BACKEND: {TYPE: "local"},
    }
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)

    assert model.config_obj.adapter.pretrained_adapter_weights
    assert model.config_obj.adapter.pretrained_adapter_weights == weights

    model.prepare_for_training()
    assert not isinstance(model.model, PreTrainedModel)
    assert isinstance(model.model, PeftModel)

    config_obj = ModelConfig.from_dict(config)
    assert config_obj.input_features[0].preprocessing.max_sequence_length is None
    assert config_obj.output_features[0].preprocessing.max_sequence_length is None


def _compare_models(model_1: torch.nn.Module, model_2: torch.nn.Module) -> bool:
    # For a full explanation of this 8-bit workaround, see https://github.com/ludwig-ai/ludwig/pull/3606

    # TODO: Uncomment "filter_for_weight_format()" method definition and enable its usage once GPU tests are set up.
    # def filter_for_weight_format(i):
    #     """Remove bitsandbytes metadata keys added on state dict creation.
    #
    #     8-bit quantized models that have been put on gpu will have a set of `weight_format` keys in their state dict.
    #     These contain strings that are used to reshape quantized tensors, however these have no impact until the state
    #     dict is loaded into a model. These keys were causing `torch.equal` to raise an exception, so we skip them in
    #     the evaluation.
    #     """
    #     return "weight_format" not in i[0]

    # model_1_filtered_state_dict = filter(filter_for_weight_format, model_1.state_dict().items())
    # model_2_filtered_state_dict = filter(filter_for_weight_format, model_2.state_dict().items())

    # Source: https://discuss.pytorch.org/t/check-if-models-have-same-weights/4351/6

    if model_1.__class__.__name__ != model_2.__class__.__name__:
        return False

    if (
        hasattr(model_1, "model")
        and hasattr(model_2, "model")
        and not _compare_models(model_1=model_1.model, model_2=model_2.model)
    ):
        return False

    for key_item_1, key_item_2 in zip(model_1.state_dict().items(), model_2.state_dict().items()):
        if not torch.equal(key_item_1[1], key_item_2[1]):
            return False

    return True


def test_global_max_sequence_length_for_llms():
    """Ensures that user specified global_max_sequence_length can never be greater than the model's context
    length."""
    config = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: "HuggingFaceH4/tiny-random-LlamaForCausalLM",
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
    }
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)

    # Default value is set based on model's context_len
    assert model.global_max_sequence_length == 2048

    # Override to a larger value in the config
    config["preprocessing"] = {"global_max_sequence_length": 4096}
    config_obj = ModelConfig.from_dict(config)
    model = LLM(config_obj)

    # Check that the value can never be larger than the model's context_len
    assert model.global_max_sequence_length == 2048


def test_local_path_loading(tmpdir):
    """Tests that local paths can be used to load models."""

    from huggingface_hub import snapshot_download

    # Download the model to a local directory
    local_path: str = f"{str(tmpdir)}/test_local_path_loading"
    repo_id: str = "HuggingFaceH4/tiny-random-LlamaForCausalLM"
    os.makedirs(local_path, exist_ok=True)
    snapshot_download(repo_id=repo_id, local_dir=local_path)

    # Load the model using the local path
    config1 = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: local_path,
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
    }
    config_obj1 = ModelConfig.from_dict(config1)
    model1 = LLM(config_obj1)

    # Load the model using the repo id
    config2 = {
        MODEL_TYPE: MODEL_LLM,
        BASE_MODEL: repo_id,
        INPUT_FEATURES: [text_feature(name="input", encoder={"type": "passthrough"})],
        OUTPUT_FEATURES: [text_feature(name="output")],
    }
    config_obj2 = ModelConfig.from_dict(config2)
    model2 = LLM(config_obj2)

    # Check that the models are the same
    assert _compare_models(model1.model, model2.model)



import copy
import random
from collections import deque, namedtuple
from typing import Any, Deque, Dict, List, Tuple, Union

import pandas as pd

from ludwig.config_sampling.parameter_sampling import handle_property_type, ParameterBaseTypes
from ludwig.constants import SEQUENCE, TEXT, TIMESERIES
from ludwig.data.dataset_synthesizer import build_synthetic_dataset_df
from ludwig.schema.model_types.base import ModelConfig
from ludwig.types import ModelConfigDict
from ludwig.utils.misc_utils import merge_dict

# number of examples to generate for synthetic dataset
NUM_SYNTHETIC_EXAMPLES = 10

ConfigOption = namedtuple("ConfigOption", ["config_option", "fully_explored"])


def explore_properties(
    jsonschema_properties: Dict[str, Any],
    parent_parameter_path: str,
    dq: Deque[ConfigOption],
    allow_list: List[str] = [],
) -> Deque[Tuple[Dict, bool]]:
    """Recursively explores the `properties` part of any subsection of the schema.

    Args:
        jsonschema_properties: any properties section of the schema.
        parent_parameter_path: period-delimited list of parent dictionary keys up to the given jsonschema_properties
            (e.g. defaults.number.preprocessing)
        dq: dequeue data structure that stores tuples of (config_options, fully_explored).
            config_options: Dict[str, List], fully_explored: bool is a dictionary is a dictionary of parameter name to
            list of values to explore.
            fully_explored is a boolean value indicating whether all subsections of the properties dictionary have been
            explored.
        allow_list: list of top level keys of the properties dictionary to skip.

    Returns:
        A deque of (dict, bool) tuples.
        - The first element of the tuple contains a dictionary of config options, which maps from a ludwig
            config parameter to a list of the values to be explored for that parameter. Here's an example:
                trainer.batch_size: ["auto", 2, 43]
                trainer.learning_rate: ["auto", 0.1, 0.00002, 0.32424]
                ...
        - The second element of the tuple is whether we've explored this "config path"
            fully. This is important to track when recursing into nested structures.
    """
    # processed_dq will contain complete config options with all the parameters in the properties dictionary
    # dq will contain configs options that are still being completed.
    processed_dq = deque()
    while dq and not dq[0].fully_explored:
        for parameter_name_or_section, jsonschema_property in jsonschema_properties.items():
            if allow_list and parameter_name_or_section not in allow_list:
                continue

            parameter_path = (
                f"{parent_parameter_path}.{parameter_name_or_section}"
                if parent_parameter_path
                else parameter_name_or_section
            )
            config_options, _ = dq.popleft()

            if "properties" in jsonschema_property and "allOf" in jsonschema_property:
                for child_item in jsonschema_property["allOf"]:
                    expanded_config_options_dq = explore_from_all_of(
                        config_options=copy.deepcopy(config_options), item=child_item, key_so_far=parameter_path
                    )
                    # add returned child config options to the deque to be processed.
                    dq.extend(expanded_config_options_dq)

            elif "properties" in jsonschema_property and "allOf" not in jsonschema_property:
                # This is the case where we don't have a list of properties, just a properties
                # dictionary nested inside another.
                child_properties = jsonschema_property["properties"]
                # a new dequeue to be passed to explore parameters from
                raw_entry = deque([ConfigOption(copy.deepcopy(config_options), False)])
                child_config_options_dq = explore_properties(child_properties, parameter_path, raw_entry)
                merged_config_options_dq = merge_dq(config_options, child_config_options_dq)
                # add returned config options to the deque to be processed.
                dq.extend(merged_config_options_dq)

            else:
                # this is the base case.
                parameter_samples = get_samples(jsonschema_property)
                if parameter_samples:
                    config_options[parameter_path] = parameter_samples

                # add config_options back to queue. fully_explored = False because we still didn't finish
                # exploring all the keys in the properties dictionary.
                dq.appendleft(ConfigOption(config_options, False))

        # at this point, we finished exploring all keys of the properties dictionary. Add all config options
        # to the processed queue.
        while dq:
            config_options, _ = dq.popleft()
            processed_dq.append(ConfigOption(config_options, True))

    return processed_dq


def get_samples(jsonschema_property: Dict[str, Any]) -> List[ParameterBaseTypes]:
    """Get possible values for a leaf property (no sub-properties).

    Args:
        jsonschema_property: leaf property in the schema. Has no sub-properties.
    """
    if "oneOf" in jsonschema_property:
        temp = []
        for elem in jsonschema_property["oneOf"]:
            temp += get_potential_values(elem)
        return temp
    else:
        return get_potential_values(jsonschema_property)


def merge_dq(config_options: Dict[str, Any], child_config_options_dq: Deque[ConfigOption]) -> Deque[ConfigOption]:
    """Merge config_options with the child_config_options in the dq."""
    dq = deque()
    while child_config_options_dq:
        child_config_options, visited = child_config_options_dq.popleft()
        cfg = merge_dict(child_config_options, config_options)
        dq.append(ConfigOption(cfg, visited))
    return dq


def explore_from_all_of(config_options: Dict[str, Any], item: Dict[str, Any], key_so_far: str) -> Deque[ConfigOption]:
    """Takes a child of `allOf` and calls `explore_properties` on it."""
    for parameter_name_or_section in item["if"]["properties"]:
        config_options[key_so_far + "." + parameter_name_or_section] = item["if"]["properties"][
            parameter_name_or_section
        ]["const"]
    jsonschema_properties = item["then"]["properties"]
    raw_entry = deque([ConfigOption(copy.deepcopy(config_options), False)])
    return explore_properties(jsonschema_properties, parent_parameter_path=key_so_far, dq=raw_entry)


def get_potential_values(item: Dict[str, Any]) -> List[Union[ParameterBaseTypes, List[ParameterBaseTypes]]]:
    """Returns a list of values to explore for a config parameter.

    Param:
        item: config parameter-specific dictionary. Considered as a leaf in the schema. Contains type, default, and
            parameter metadata, etc.
    """
    temp = []
    # Case where we're using OneOf (e.g. to allow batch size 'auto' and integers)
    if isinstance(item["type"], list):
        for property_type in item["type"]:
            temp += handle_property_type(property_type, item)
    else:
        temp += handle_property_type(item["type"], item)

    # Make sure values are unique. Not using set because some values are unhashable.
    unique_temp = []
    for temp_item in temp:
        if temp_item not in unique_temp:
            unique_temp.append(temp_item)
    return unique_temp


def generate_possible_configs(config_options: Dict[str, Any]):
    """Generate exhaustive configs from config_options.

    This function does not take a cross product of all the options for all the config parameters. It selects parameter
    values independently from each other.

    Args:
        config_options: dictionary mapping from ludwig config parameter to all values to be explored.
            Here's an example of what it could look like:

                trainer.batch_size: ["auto", 2, 43]
                trainer.learning_rate: ["auto", 0.1, 0.00002, 0.32424]
                ...
    """
    # The number of configs to generate is the max length of the lists of samples over all parameters.
    num_configs = 1
    for parameter_name in config_options:
        if isinstance(config_options[parameter_name], list):
            num_configs = max(num_configs, len(config_options[parameter_name]))
            config_options[parameter_name] = deque(config_options[parameter_name])

    for _ in range(num_configs):
        config = {}
        for parameter_name in config_options:
            # if parameter is regular parameter with explored values.
            if config_options[parameter_name] and not isinstance(config_options[parameter_name], str):
                config[parameter_name] = config_options[parameter_name].popleft()
            # case for parameters where we don't have choices such as `encoder.type: parallel_cnn` that
            # cause the downstream parameters to change.
            elif isinstance(config_options[parameter_name], str):
                config[parameter_name] = config_options[parameter_name]
        yield create_nested_dict(config)


def create_nested_dict(flat_dict: Dict[str, Union[float, str]]) -> ModelConfigDict:
    """Generate a nested dict out of a flat dict whose keys are delimited by a delimiter character.

    Args:
        flat_dict: potential generated baseline config. Here's an example of what it could look like:

            trainer.batch_size: 324
            trainer.learning_rate: 0.0635

        The expected output would be

            trainer:
                batch_size: 324
                learning_rate: 0.0635
    """

    def to_nested_format(parameter_name: str, value: Union[str, int, float], delimiter: str = ".") -> Dict[str, Any]:
        # https://stackoverflow.com/a/40401961
        split_parameter_name = parameter_name.split(delimiter)
        for parameter_name_or_section in reversed(split_parameter_name):
            value = {parameter_name_or_section: value}
        return value

    config = {}
    for parameter_name_or_section in flat_dict:
        config = merge_dict(
            config, to_nested_format(parameter_name_or_section, copy.deepcopy(flat_dict[parameter_name_or_section]))
        )
    return config


def combine_configs(
    explored: Deque[Tuple[Dict, bool]], config: ModelConfigDict
) -> List[Tuple[ModelConfigDict, pd.DataFrame]]:
    """Merge base config with explored sections.

    Args:
        explored: deque containing all the config options.
        config: base Ludwig config to merge the explored configs with.
    """
    dataset = build_synthetic_dataset_df(NUM_SYNTHETIC_EXAMPLES, config)
    ret = []
    for config_options, _ in explored:
        for default_config in generate_possible_configs(config_options=config_options):
            merged_config = merge_dict(copy.deepcopy(config), default_config)
            try:
                ModelConfig.from_dict(merged_config)
                ret.append((merged_config, dataset))
            except Exception:
                pass
    return ret


def combine_configs_for_comparator_combiner(
    explored: Deque[Tuple], config: ModelConfigDict
) -> List[Tuple[ModelConfigDict, pd.DataFrame]]:
    """Merge base config with explored sections.

    Completes the entity_1 and entity_2 paramters of the comparator combiner.

    Args:
        explored: deque containing all the config options.
        config: base Ludwig config to merge the explored configs with.
    """
    dataset = build_synthetic_dataset_df(NUM_SYNTHETIC_EXAMPLES, config)
    ret = []
    for item in explored:
        for default_config in generate_possible_configs(config_options=item[0]):
            merged_config = merge_dict(copy.deepcopy(config), default_config)

            # create two random lists for entity1 and entity2
            entity_names = [feature["name"] for feature in config["input_features"]]
            random.shuffle(entity_names)
            entity_1_size = random.randint(1, len(entity_names) - 1)
            merged_config["combiner"]["entity_1"] = entity_names[:entity_1_size]
            merged_config["combiner"]["entity_2"] = entity_names[entity_1_size:]
            try:
                ModelConfig.from_dict(merged_config)
                ret.append((merged_config, dataset))
            except Exception:
                pass
    return ret


def combine_configs_for_sequence_combiner(
    explored: Deque[Tuple], config: ModelConfigDict
) -> List[Tuple[ModelConfigDict, pd.DataFrame]]:
    """Merge base config with explored sections.

    Uses the right reduce_output strategy for the sequence and sequence_concat combiners.

    Args:
        explored: deque containing all the config options.
        config: base Ludwig config to merge the explored configs with.
    """
    dataset = build_synthetic_dataset_df(NUM_SYNTHETIC_EXAMPLES, config)
    ret = []
    for item in explored:
        for default_config in generate_possible_configs(config_options=item[0]):
            merged_config = merge_dict(copy.deepcopy(config), default_config)
            for i in range(len(merged_config["input_features"])):
                if merged_config["input_features"][i]["type"] in {SEQUENCE, TEXT, TIMESERIES}:
                    merged_config["input_features"][0]["encoder"] = {"type": "embed", "reduce_output": None}
            try:
                ModelConfig.from_dict(merged_config)
                ret.append((merged_config, dataset))
            except Exception:
                pass
    return ret



import random
from typing import Any, Dict, List, Union

from ludwig.schema.metadata.parameter_metadata import ExpectedImpact

# base types for ludwig config parameters.
ParameterBaseTypes = Union[str, float, int, bool, None]


def handle_property_type(
    property_type: str, item: Dict[str, Any], expected_impact: ExpectedImpact = ExpectedImpact.HIGH
) -> List[Union[ParameterBaseTypes, List[ParameterBaseTypes]]]:
    """Return possible parameter values for a parameter type.

    Args:
        property_type: type of the parameter (e.g. array, number, etc.)
        item: dictionary containing details on the parameter such as default, min and max values.
        expected_impact: threshold expected impact that we'd like to include.
    """
    parameter_metadata = item.get("parameter_metadata", None)
    if not parameter_metadata:
        return []

    # don't explore internal only parameters.
    if parameter_metadata.get("internal_only", True):
        return []

    # don't explore parameters that have expected impact less than HIGH.
    if parameter_metadata.get("expected_impact", ExpectedImpact.LOW) < expected_impact:
        return []

    if property_type == "number":
        return explore_number(item)
    elif property_type == "integer":
        return explore_integer(item)
    elif property_type == "string":
        return explore_string(item)
    elif property_type == "boolean":
        return explore_boolean()
    elif property_type == "null":
        return explore_null()
    elif property_type == "array":
        return explore_array(item)
    else:
        return []


def explore_array(item: Dict[str, Any]) -> List[List[ParameterBaseTypes]]:
    """Return possible parameter values for the `array` parameter type.

    Args:
        item: dictionary containing details on the parameter such as default, min and max values.
    """

    candidates = []
    if "default" in item and item["default"]:
        candidates.append(item["default"])

    item_choices = []
    maxlen = 0

    # In the case where the length of the array isn't defined.
    if not isinstance(item["items"], list):
        return []

    for item_of in item["items"]:
        choices = handle_property_type(item_of["type"], item_of)
        maxlen = max(maxlen, len(choices))
        item_choices.append(choices)

    # pad to same length
    for i in range(len(item_choices)):
        item_choices[i] = maxlen * item_choices[i]
        item_choices[i] = item_choices[i][:maxlen]

    merged = list(zip(*item_choices)) + candidates
    return [list(tup) for tup in merged]


def explore_number(item: Dict[str, Any]) -> List[ParameterBaseTypes]:
    """Return possible parameter values for the `number` parameter type.

    Args:
        item: dictionary containing details on the parameter such as default, min and max values.
    TODO(Wael): Improve logic.
    """
    minimum, maximum = 0, 1
    if "default" not in item or item["default"] is None:
        candidates = []
    else:
        candidates = [1, 2, item["default"], 2 * (item["default"] + 1), item["default"] // 2, -1 * item["default"]]

    if "minimum" in item:
        minimum = item["minimum"]
        candidates = [num for num in candidates if num > minimum]
    if "maximum" in item:
        maximum = item["maximum"]
        candidates = [num for num in candidates if num < maximum]
    return candidates + [random.random() * 0.99 * maximum]


def explore_integer(item: Dict[str, Any]) -> List[ParameterBaseTypes]:
    """Return possible parameter values for the `integer` parameter type.

    Args:
        item: dictionary containing details on the parameter such as default, min and max values.
    TODO(Wael): Improve logic.
    """
    minimum, maximum = 0, 10

    if "default" not in item or item["default"] is None:
        candidates = []
    else:
        candidates = [item["default"], 2 * (item["default"] + 1), item["default"] // 2, -1 * item["default"]]

    if "minimum" in item:
        minimum = item["minimum"]
        candidates = [num for num in candidates if num >= item["minimum"]]
    if "maximum" in item:
        maximum = item["maximum"]
        candidates = [num for num in candidates if num <= item["maximum"]]

    return candidates + [random.randint(minimum, maximum)]


def explore_string(item: Dict[str, Any]) -> List[ParameterBaseTypes]:
    """Return possible parameter values for the `string` parameter type.

    Args:
        item: dictionary containing details on the parameter such as default, min and max values.
    """

    if "enum" in item:
        return item["enum"]
    return [item["default"]]


def explore_boolean() -> List[bool]:
    """Return possible parameter values for the `boolean` parameter type (i.e. [True, False])"""
    return [True, False]


def explore_null() -> List[None]:
    """Return possible parameter values for the `null` parameter type (i.e. [None])"""
    return [None]



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from functools import partial
from typing import Dict, List, Optional, Tuple, Union

import torch
from torch import Tensor
from transformers import PreTrainedTokenizer

from ludwig.constants import (
    COLUMN,
    IGNORE_INDEX_TOKEN_ID,
    LAST_PREDICTIONS,
    LENGTHS,
    NAME,
    PREDICTIONS,
    PREPROCESSING,
    PROBABILITIES,
    PROBABILITY,
    PROC_COLUMN,
    RESPONSE,
    TEXT,
)
from ludwig.features.base_feature import BaseFeatureMixin, OutputFeature
from ludwig.features.feature_utils import compute_sequence_probability, compute_token_probabilities
from ludwig.features.sequence_feature import (
    _SequencePostprocessing,
    _SequencePreprocessing,
    SequenceInputFeature,
    SequenceOutputFeature,
)
from ludwig.modules.metric_registry import get_metric_tensor_input
from ludwig.schema.features.text_feature import TextInputFeatureConfig, TextOutputFeatureConfig
from ludwig.types import FeatureMetadataDict, PreprocessingConfigDict, TrainingSetMetadataDict
from ludwig.utils.math_utils import softmax
from ludwig.utils.strings_utils import (
    build_sequence_matrix,
    create_vocabulary,
    get_tokenizer,
    SpecialSymbol,
    UNKNOWN_SYMBOL,
    Vocabulary,
)

logger = logging.getLogger(__name__)


def get_decoded_targets_and_predictions(
    targets: Tensor,
    predictions: Dict[str, Tensor],
    tokenizer: PreTrainedTokenizer,
) -> Tuple[List[str], List[str]]:
    """Returns the decoded targets and predictions, accounting for IGNORE_INDEX_TOKEN_ID."""
    sanitized_targets = torch.where(targets != IGNORE_INDEX_TOKEN_ID, targets, tokenizer.pad_token_id)
    sanitized_predictions = torch.where(
        predictions[PREDICTIONS] != IGNORE_INDEX_TOKEN_ID,
        predictions[PREDICTIONS],
        tokenizer.pad_token_id,
    )
    decoded_targets = tokenizer.batch_decode(sanitized_targets, skip_special_tokens=True)
    decoded_predictions = tokenizer.batch_decode(sanitized_predictions, skip_special_tokens=True)
    return decoded_targets, decoded_predictions


class TextFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return TEXT

    @staticmethod
    def cast_column(column, backend):
        return column.astype(str)

    @staticmethod
    def feature_meta(column, preprocessing_parameters: PreprocessingConfigDict, backend) -> Vocabulary:
        return create_vocabulary(
            column,
            tokenizer_type=preprocessing_parameters["tokenizer"],
            num_most_frequent=preprocessing_parameters["most_common"],
            lowercase=preprocessing_parameters["lowercase"],
            vocab_file=preprocessing_parameters["vocab_file"],
            unknown_symbol=preprocessing_parameters["unknown_symbol"],
            padding_symbol=preprocessing_parameters["padding_symbol"],
            pretrained_model_name_or_path=preprocessing_parameters["pretrained_model_name_or_path"],
            ngram_size=preprocessing_parameters["ngram_size"],
            compute_idf=preprocessing_parameters["compute_idf"],
            processor=backend.df_engine,
        )

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        vocabulary = TextFeatureMixin.feature_meta(column, preprocessing_parameters, backend)
        logger.info(
            f"Max length of feature '{column.name}': {vocabulary.line_length_max} (without start and stop symbols)"
        )

        # Use sequence_length if provided, otherwise use max length found in dataset.
        if preprocessing_parameters["sequence_length"] is not None:
            logger.info(
                f"Setting max length to sequence_length={preprocessing_parameters['sequence_length']} provided in "
                f"preprocessing parameters"
            )
            max_sequence_length = preprocessing_parameters["sequence_length"]
            max_sequence_length_99ptile = preprocessing_parameters["sequence_length"]
        else:
            max_sequence_length = vocabulary.line_length_max + 2  # For start and stop symbols.
            max_sequence_length_99ptile = vocabulary.line_length_99ptile + 2  # For start and stop symbols.
            logger.info(f"Setting max length using dataset: {max_sequence_length} (including start and stop symbols)")

            # If max_sequence_length is None, then use the max length found in the dataset.
            if (
                preprocessing_parameters["max_sequence_length"] is not None
                and preprocessing_parameters["max_sequence_length"] < max_sequence_length
            ):
                logger.info(
                    f"Truncating max length with max_sequence_length={preprocessing_parameters['max_sequence_length']} "
                    f"from preprocessing parameters"
                )
                max_sequence_length = preprocessing_parameters["max_sequence_length"]
                max_sequence_length_99ptile = min(vocabulary.line_length_99ptile, max_sequence_length)

        logger.info(f"max sequence length is {max_sequence_length} for feature '{column.name}'")

        return {
            "idx2str": vocabulary.vocab,
            "str2idx": vocabulary.str2idx,
            "str2freq": vocabulary.str2freq,
            "str2idf": vocabulary.str2idf,
            "vocab_size": len(vocabulary.vocab),
            "max_sequence_length": max_sequence_length,
            "max_sequence_length_99ptile": max_sequence_length_99ptile,
            "pad_idx": vocabulary.pad_idx,
            "padding_symbol": vocabulary.padding_symbol,
            "unknown_symbol": vocabulary.unknown_symbol,
        }

    @staticmethod
    def feature_data(column, metadata, preprocessing_parameters: PreprocessingConfigDict, backend):
        # TODO(1891): Remove backward compatibility hack once all models have been retrained with Ludwig after
        # https://github.com/ludwig-ai/ludwig/pull/1859.
        prefix = ""
        padding_symbol_metadata_key = "padding_symbol"
        unknown_symbol_metadata_key = "unknown_symbol"
        if "str2idx" not in metadata:
            prefix = "word_"
            padding_symbol_metadata_key = "word_pad_symbol"
            unknown_symbol_metadata_key = "word_unk_symbol"

        # ensure preprocessing param values match the metadata determined from dataset
        preprocessing_parameters["padding_symbol"] = metadata[padding_symbol_metadata_key]
        preprocessing_parameters["unknown_symbol"] = metadata[unknown_symbol_metadata_key]
        if preprocessing_parameters["fill_value"] == UNKNOWN_SYMBOL:
            preprocessing_parameters["fill_value"] = preprocessing_parameters["unknown_symbol"]
        if (
            "computed_fill_value" in preprocessing_parameters
            and preprocessing_parameters["computed_fill_value"] == UNKNOWN_SYMBOL
        ):
            preprocessing_parameters["computed_fill_value"] = preprocessing_parameters["unknown_symbol"]

        sequences = column

        return build_sequence_matrix(
            sequences=sequences,
            inverse_vocabulary=metadata[f"{prefix}str2idx"],
            tokenizer_type=preprocessing_parameters[f"{prefix}tokenizer"],
            length_limit=metadata[f"{prefix}max_sequence_length"],
            padding_symbol=metadata[padding_symbol_metadata_key],
            padding=preprocessing_parameters["padding"],
            unknown_symbol=metadata[unknown_symbol_metadata_key],
            lowercase=preprocessing_parameters["lowercase"],
            tokenizer_vocab_file=preprocessing_parameters[f"{prefix}vocab_file"],
            pretrained_model_name_or_path=preprocessing_parameters["pretrained_model_name_or_path"],
            processor=backend.df_engine,
        )

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        proc_df[feature_config[PROC_COLUMN]] = TextFeatureMixin.feature_data(
            input_df[feature_config[COLUMN]],
            metadata[feature_config[NAME]],
            preprocessing_parameters,
            backend,
        )
        return proc_df


class TextInputFeature(TextFeatureMixin, SequenceInputFeature):
    def __init__(self, input_feature_config: TextInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, encoder_obj=encoder_obj, **kwargs)

    def forward(self, inputs, mask=None):
        assert isinstance(inputs, torch.Tensor)
        assert (
            inputs.dtype == torch.int8
            or inputs.dtype == torch.int16
            or inputs.dtype == torch.int32
            or inputs.dtype == torch.int64
        )
        assert len(inputs.shape) == 2

        inputs_mask = torch.not_equal(inputs, SpecialSymbol.PADDING.value)

        inputs_exp = inputs.type(torch.int32)
        lengths = torch.sum(inputs_mask.type(torch.int32), dim=1)
        encoder_output = self.encoder_obj(inputs_exp, mask=inputs_mask)
        encoder_output[LENGTHS] = lengths

        return encoder_output

    @property
    def input_dtype(self):
        return torch.int32

    @property
    def input_shape(self):
        return torch.Size([self.encoder_obj.config.max_sequence_length])

    def update_config_after_module_init(self, feature_config):
        feature_config.encoder = self.encoder_obj.config

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.encoder.vocab = feature_metadata["idx2str"]
        feature_config.encoder.vocab_size = len(feature_metadata["idx2str"])
        feature_config.encoder.max_sequence_length = feature_metadata["max_sequence_length"]
        feature_config.encoder.pad_idx = feature_metadata["pad_idx"]
        feature_config.encoder.num_tokens = len(feature_metadata["idx2str"])
        feature_config.encoder.str2freq = feature_metadata["str2freq"]
        feature_config.encoder.str2idf = feature_metadata["str2idf"]
        feature_config.encoder.skip = feature_metadata[PREPROCESSING].get("cache_encoder_embeddings", False)

    @staticmethod
    def get_schema_cls():
        return TextInputFeatureConfig

    @property
    def output_shape(self) -> torch.Size:
        return self.encoder_obj.output_shape

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SequencePreprocessing(metadata)


class TextOutputFeature(TextFeatureMixin, SequenceOutputFeature):
    def __init__(
        self,
        output_feature_config: Union[TextOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        super().__init__(output_feature_config, output_features, **kwargs)

    @classmethod
    def get_output_dtype(cls):
        return torch.int32

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.decoder_obj.config.max_sequence_length])

    def update_metrics(
        self,
        targets: Tensor,
        predictions: Dict[str, Tensor],
        tokenizer: Optional[PreTrainedTokenizer] = None,
    ) -> None:
        """Updates metrics with the given targets and predictions.

        If decoded_targets and decoded_predictions are provided, as through LLM model types, then additional
        response-based metrics like BLEU and ROUGE are also computed.

        Args:
            targets: Tensor with target values for this output feature.
            predictions: Dict of tensors returned by predictions().
        """
        if tokenizer is not None:
            # Decode the targets and predictions to compute response-based metrics using the initialized tokenizer.
            decoded_targets, decoded_predictions = get_decoded_targets_and_predictions(targets, predictions, tokenizer)

        for metric_name, metric_fn in self._metric_functions.items():
            prediction_key = get_metric_tensor_input(metric_name)
            try:
                if prediction_key == RESPONSE:
                    if tokenizer is not None:
                        # RESPONSE metrics cannot be computed if decoded texts are not provided.
                        # Decoded texts are only provided using the LLM model type.
                        if decoded_targets is not None and decoded_predictions is not None:
                            # Move metric function to the device of the predictions.
                            # For CUDA, it can be computed on any of the GPUs since it uses allgather to collect
                            # the results from all GPUs and compute the final metric.
                            # We use 'predictions' as the key since it is always present in the predictions dict.
                            device = "cuda" if predictions["predictions"].is_cuda else "cpu"
                            metric_fn = metric_fn.to(device)
                            if metric_name == "bleu":
                                # BLEU takes in targets as a list.
                                metric_fn.update(decoded_predictions, [decoded_targets])
                            else:
                                metric_fn.update(decoded_predictions, decoded_targets)
                else:
                    metric_fn = metric_fn.to(predictions[prediction_key].device)
                    metric_fn.update(predictions[prediction_key].detach(), targets)
            except Exception as e:
                logger.info(f"Ran into error when calculating metric {metric_name}. Skipping. The error is: {e}")

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.decoder.vocab_size = feature_metadata["vocab_size"]
        feature_config.decoder.max_sequence_length = feature_metadata["max_sequence_length"]
        if isinstance(feature_config.loss.class_weights, (list, tuple)):
            # [0, 0] for UNK and PAD
            feature_config.loss.class_weights = [0, 0] + feature_config.loss.class_weights
            if len(feature_config.loss.class_weights) != feature_config.decoder.vocab_size:
                raise ValueError(
                    "The length of class_weights ({}) is not compatible with "
                    "the number of classes ({})".format(
                        len(feature_config.loss.class_weights), feature_config.decoder.vocab_size
                    )
                )

        if isinstance(feature_config.loss.class_weights, dict):
            if feature_metadata["str2idx"].keys() != feature_config.loss.class_weights.keys():
                raise ValueError(
                    "The class_weights keys ({}) are not compatible with "
                    "the classes ({}) of feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and consider there needs to be a weight "
                    "for the <UNK> class too.".format(
                        feature_config.loss.class_weights.keys(),
                        feature_metadata["str2idx"].keys(),
                        feature_config.column,
                    )
                )
            else:
                class_weights = feature_config.loss.class_weights
                idx2str = feature_metadata["idx2str"]
                class_weights_list = [class_weights[s] for s in idx2str]
                feature_config.loss.class_weights = class_weights_list

        if feature_config.loss.class_similarities_temperature > 0:
            if feature_config.class_similarities:
                distances = feature_config.class_similarities
                temperature = feature_config.loss.class_similarities_temperature
                for i in range(len(distances)):
                    distances[i, :] = softmax(distances[i, :], temperature=temperature)
                feature_config.loss.class_similarities = distances
            else:
                raise ValueError(
                    "class_similarities_temperature > 0,"
                    "but no class similarities are provided "
                    "for feature {}".format(feature_config.column)
                )

    @staticmethod
    def calculate_overall_stats(
        predictions,
        targets,
        train_set_metadata,
    ):
        return {}

    def postprocess_predictions(
        self,
        result,
        metadata,
    ):
        # todo: refactor to reuse SequenceOutputFeature.postprocess_predictions
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"

        tokenizer = None
        if metadata["preprocessing"]["tokenizer"] == "hf_tokenizer":
            tokenizer = get_tokenizer(
                metadata["preprocessing"]["tokenizer"],
                metadata["preprocessing"]["vocab_file"],
                metadata["preprocessing"]["pretrained_model_name_or_path"],
            )

        if predictions_col in result:
            token_col = result[predictions_col]

            def idx2str(pred):
                if tokenizer is None:
                    return [
                        metadata["idx2str"][token] if token < len(metadata["idx2str"]) else UNKNOWN_SYMBOL
                        for token in pred
                    ]
                return tokenizer.tokenizer.batch_decode(pred, skip_special_tokens=True)

            result[predictions_col] = token_col.map(idx2str)

            # Add additional response column that represents the predicted text output
            # as a single string instead of a list of tokens.
            def idx2response(pred):
                if tokenizer is None:
                    # This works because we treat each word as a token.
                    return " ".join(
                        [
                            metadata["idx2str"][token] if token < len(metadata["idx2str"]) else UNKNOWN_SYMBOL
                            for token in pred
                        ]
                    )
                return tokenizer.tokenizer.batch_decode([pred], skip_special_tokens=True)

            result[f"{self.feature_name}_response"] = token_col.map(idx2response)

        last_preds_col = f"{self.feature_name}_{LAST_PREDICTIONS}"
        if last_preds_col in result:

            def last_idx2str(last_pred):
                if last_pred < len(metadata["idx2str"]):
                    return metadata["idx2str"][last_pred]
                return UNKNOWN_SYMBOL

            result[last_preds_col] = result[last_preds_col].map(last_idx2str)

        probs_col = f"{self.feature_name}_{PROBABILITIES}"
        prob_col = f"{self.feature_name}_{PROBABILITY}"

        # "Summarizes" the `result`'s probability-related output:
        # - result[probs_col]:
        #       Each row is now a list of "max" probabilities. Each element is the probability of the argmax token for
        #       the given time step.
        #
        #       Note that we intentionally do not return full list of probabilties for each time step because the output
        #       of postprocess_predictions is saved to disk and the full probability distribution can be huge,
        #       especially for large vocab sizes:
        #           dataset_size x sequence_length x vocab_size
        #
        #       TODO: Add a mechanism that lets the user save the full probability distribution if they want.
        # - result[prob_col]:
        #       Each row is the overall probability of the sequence. This is the product of the max probabilities over
        #       all time steps.
        if probs_col in result:
            # result[probs_col]: From PredictModule, each row has a list of size (sequence_length) of a list of
            # probabiltiies of (vocab_size). compute_token_probabilities gets the maximum probability per timestep.
            result[probs_col] = result[probs_col].map(compute_token_probabilities)
            result[prob_col] = result[probs_col].map(
                partial(
                    compute_sequence_probability,
                    max_sequence_length=metadata["max_sequence_length"],
                    return_log_prob=True,
                ),
            )

        lengths_col = f"{self.feature_name}_{LENGTHS}"
        if lengths_col in result:
            del result[lengths_col]

        return result

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _SequencePostprocessing(metadata)

    @staticmethod
    def get_schema_cls():
        return TextOutputFeatureConfig



# Source: https://github.com/Raschka-research-group/coral-pytorch/blob/main/coral_pytorch/losses.py
# Sebastian Raschka 2020-2021
# coral_pytorch
# Author: Sebastian Raschka <sebastianraschka.com>
#
# License: MIT

import torch
import torch.nn.functional as F


def corn_loss(logits, y_train, num_classes):
    """Computes the CORN loss described in our forthcoming 'Deep Neural Networks for Rank Consistent Ordinal
    Regression based on Conditional Probabilities' manuscript.

    Parameters
    ----------
    logits : torch.tensor, shape=(num_examples, num_classes-1)
        Outputs of the CORN layer.

    y_train : torch.tensor, shape=(num_examples)
        Torch tensor containing the class labels.

    num_classes : int
        Number of unique class labels (class labels should start at 0).

    Returns
    ----------
        loss : torch.tensor
        A torch.tensor containing a single loss value.

    Examples
    ----------
    >>> # Consider 8 training examples
    >>> _  = torch.manual_seed(123)
    >>> X_train = torch.rand(8, 99)
    >>> y_train = torch.tensor([0, 1, 2, 2, 2, 3, 4, 4])
    >>> NUM_CLASSES = 5
    >>> #
    >>> #
    >>> # def __init__(self):
    >>> corn_net = torch.nn.Linear(99, NUM_CLASSES-1)
    >>> #
    >>> #
    >>> # def forward(self, X_train):
    >>> logits = corn_net(X_train)
    >>> logits.shape
    torch.Size([8, 4])
    >>> corn_loss(logits, y_train, NUM_CLASSES)
    tensor(0.7127, grad_fn=<DivBackward0>)
    """
    sets = []
    for i in range(num_classes - 1):
        label_mask = y_train > i - 1
        label_tensor = (y_train[label_mask] > i).to(torch.int64)
        sets.append((label_mask, label_tensor))

    num_examples = 0
    losses = 0.0
    for task_index, s in enumerate(sets):
        train_examples = s[0]
        train_labels = s[1]

        if len(train_labels) < 1:
            continue

        num_examples += len(train_labels)
        pred = logits[train_examples, task_index]

        loss = -torch.sum(F.logsigmoid(pred) * train_labels + (F.logsigmoid(pred) - pred) * (1 - train_labels))
        losses += loss

    return losses / num_examples



import os
from dataclasses import field

from marshmallow import fields, ValidationError
from transformers import AutoConfig

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BASE_MODEL
from ludwig.error import ConfigValidationError
from ludwig.schema.metadata import LLM_METADATA
from ludwig.schema.metadata.parameter_metadata import convert_metadata_to_json

# Maps a preset LLM name to the full slash-delimited HF path. If the user chooses a preset LLM, the preset LLM name is
# replaced with the full slash-delimited HF path using this map, after JSON validation but before config object
# initialization.
MODEL_PRESETS = {
    "llama-2-7b": "meta-llama/Llama-2-7b-hf",
    "llama-2-13b": "meta-llama/Llama-2-13b-hf",
    "llama-2-70b": "meta-llama/Llama-2-70b-hf",
    "llama-2-7b-chat": "meta-llama/Llama-2-7b-chat-hf",
    "llama-2-13b-chat": "meta-llama/Llama-2-13b-chat-hf",
    "llama-2-70b-chat": "meta-llama/Llama-2-70b-chat-hf",
    "vicuna-7b": "lmsys/vicuna-7b-v1.3",
    "vicuna-13b": "lmsys/vicuna-13b-v1.3",
    "bloomz-3b": "bigscience/bloomz-3b",
    "bloomz-7b1": "bigscience/bloomz-7b1",
    "opt-350m": "facebook/opt-350m",
    "opt-1.3b": "facebook/opt-1.3b",
    "opt-6.7b": "facebook/opt-6.7b",
    "gpt-neo-2.7B": "EleutherAI/gpt-neo-2.7B",
    "gpt-j-6b": "EleutherAI/gpt-j-6b",
    "pythia-2.8b": "EleutherAI/pythia-2.8b",
    "pythia-12b": "EleutherAI/pythia-12b",
    "mistral-7b": "mistralai/Mistral-7B-v0.1",
}


@DeveloperAPI
def BaseModelDataclassField():
    description = (
        "Base pretrained model to use. This can be one of the presets defined by Ludwig, a fully qualified "
        "name of a pretrained model from the HuggingFace Hub, or a path to a directory containing a "
        "pretrained model."
    )

    def validate(model_name: str):
        """Validates and upgrades the given model name to its full path, if applicable.

        If the name exists in `MODEL_PRESETS`, returns the corresponding value from the dict; otherwise checks if the
        given name (which should be a full path) exists locally or in the transformers library.
        """
        if isinstance(model_name, str):
            if model_name in MODEL_PRESETS:
                return MODEL_PRESETS[model_name]
            if os.path.isdir(model_name):
                return model_name
            try:
                AutoConfig.from_pretrained(model_name)
                return model_name
            except OSError:
                raise ConfigValidationError(
                    f"Specified base model `{model_name}` could not be loaded. If this is a private repository, make "
                    f"sure to set HUGGING_FACE_HUB_TOKEN in your environment. Check that {model_name} is a valid "
                    "pretrained CausalLM listed on huggingface or a valid local directory containing the weights for a "
                    "pretrained CausalLM from huggingface. See: "
                    "https://huggingface.co/models?pipeline_tag=text-generation&sort=downloads for a full list."
                )
        raise ValidationError(
            f"`base_model` should be a string, instead given: {model_name}. This can be a preset or any pretrained "
            "CausalLM on huggingface. See: https://huggingface.co/models?pipeline_tag=text-generation&sort=downloads"
        )

    class BaseModelField(fields.Field):
        def _serialize(self, value, attr, obj, **kwargs):
            if isinstance(value, str):
                return value
            raise ValidationError(f"Value to serialize is not a string: {value}")

        def _deserialize(self, value, attr, obj, **kwargs):
            return validate(value)

        def _jsonschema_type_mapping(self):
            return {
                "anyOf": [
                    {
                        "type": "string",
                        "enum": list(MODEL_PRESETS.keys()),
                        "description": (
                            "Pick from a set of popular LLMs of different sizes across a variety of architecture types."
                        ),
                        "title": "preset",
                        "parameter_metadata": convert_metadata_to_json(LLM_METADATA[BASE_MODEL]["_anyOf"]["preset"]),
                    },
                    {
                        "type": "string",
                        "description": "Enter the full path to a huggingface LLM.",
                        "title": "custom",
                        "parameter_metadata": convert_metadata_to_json(LLM_METADATA[BASE_MODEL]["_anyOf"]["custom"]),
                    },
                ],
                "description": description,
                "title": "base_model_options",
                "parameter_metadata": convert_metadata_to_json(LLM_METADATA[BASE_MODEL]["_meta"]),
            }

    return field(
        metadata={
            "marshmallow_field": BaseModelField(
                required=True,
                allow_none=False,
                validate=validate,
                metadata={  # TODO: extra metadata dict probably unnecessary, but currently a widespread pattern
                    "description": description,
                    "parameter_metadata": convert_metadata_to_json(LLM_METADATA[BASE_MODEL]["_meta"]),
                },
            ),
        },
        # TODO: This is an unfortunate side-effect of dataclass init order - you cannot have non-default fields follow
        # default fields, so we have to give `base_model` a fake default of `None`.
        default=None,
    )



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import asdict
from typing import Dict, Optional, Tuple, Type, TYPE_CHECKING

import torch

from ludwig.utils.misc_utils import get_from_registry
from ludwig.utils.torch_utils import LudwigModule

if TYPE_CHECKING:
    from ludwig.schema.optimizers import BaseOptimizerConfig, GradientClippingConfig


def create_clipper(gradient_clipping_config: Optional["GradientClippingConfig"]):
    from ludwig.schema.optimizers import GradientClippingConfig

    """Utility function that will convert a None-type gradient clipping config to the correct form."""
    if isinstance(gradient_clipping_config, GradientClippingConfig):
        return gradient_clipping_config
    # Return default config if provided value is None:
    return GradientClippingConfig()


def get_optimizer_class_and_kwargs(
    optimizer_config: "BaseOptimizerConfig", learning_rate: float
) -> Tuple[Type[torch.optim.Optimizer], Dict]:
    """Returns the optimizer class and kwargs for the optimizer.

    :return: Tuple of optimizer class and kwargs for the optimizer.
    """
    from ludwig.schema.optimizers import optimizer_registry

    # Get the corresponding torch optimizer class for the given config:
    optimizer_cls = get_from_registry(optimizer_config.type.lower(), optimizer_registry)[0]

    # Create a dict of parameters to be passed to torch (i.e. everything except `type`):
    cls_kwargs = {field: value for field, value in asdict(optimizer_config).items() if field != "type"}
    cls_kwargs["lr"] = learning_rate

    return optimizer_cls, cls_kwargs


def create_optimizer(
    model: LudwigModule,
    optimizer_config: "BaseOptimizerConfig",
    learning_rate: float,
) -> torch.optim.Optimizer:
    """Returns a ready-to-use torch optimizer instance based on the given optimizer config.

    :param model: Underlying Ludwig model
    :param learning_rate: Initial learning rate for the optimizer
    :param optimizer_config: Instance of `ludwig.modules.optimization_modules.BaseOptimizerConfig`.
    :return: Initialized instance of a torch optimizer.
    """
    # Make sure the optimizer is compatible with the available resources:
    if (optimizer_config.is_paged or optimizer_config.is_8bit) and (
        not torch.cuda.is_available() or torch.cuda.device_count() == 0
    ):
        raise ValueError(
            "Cannot use a paged or 8-bit optimizer on a non-GPU machine. "
            "Please use a different optimizer or run on a machine with a GPU."
        )

    optimizer_cls, optimizer_kwargs = get_optimizer_class_and_kwargs(optimizer_config, learning_rate)
    return optimizer_cls(model.parameters(), **optimizer_kwargs)



from typing import List, Set

from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata.parameter_metadata import INTERNAL_ONLY
from ludwig.schema.utils import ludwig_dataclass

"""
NOTE TO DEVELOPERS: the implementation of the schema classes below must match the parameters of the HF PretrainedConfig
class exactly. This is because we convert this object into the matching HF PretrainedConfig object before passing it to
the model. Additionally, for loading and saving pretrained models, we take the config from the existing model and load
it into this config before saving. As such, if any params needed by the pretrained model are missing, we will not be
able to load checkpoints correctly.

A common mistake is to look at the PretrainedConfig __init__ method params and ignore any additional **kwargs. In some
cases, these kwargs are used to set additional params on the config object. For example, the DebertaConfig class has
`position_buckets` as a kwarg param, but it nonetheless requires this to construct the model architecture.

To debug issues with missing parameters, try printing out the `model.config` of the pretrained transformer and check
for any params it includes that are not present in your schema config.
"""


@ludwig_dataclass
class DebertaModelParams(schema_utils.BaseMarshmallowConfig):
    @classmethod
    def get_hf_config_param_names(cls) -> Set[str]:
        return DebertaModelParams.get_valid_field_names()

    # Model architecture params for training from scratch
    # TODO(travis): conditionally disable setting these when `use_pretrained=True`.
    vocab_size: int = schema_utils.PositiveInteger(
        default=None,
        description="",
        parameter_metadata=INTERNAL_ONLY,
    )

    hidden_size: int = schema_utils.PositiveInteger(
        default=1536,
        description="Dimensionality of the encoder layers and the pooler layer.",
    )

    num_hidden_layers: int = schema_utils.PositiveInteger(
        default=24,
        description="Number of hidden layers in the Transformer encoder.",
    )

    num_attention_heads: int = schema_utils.PositiveInteger(
        default=24,
        description="Number of attention heads for each attention layer in the Transformer encoder.",
    )

    intermediate_size: int = schema_utils.PositiveInteger(
        default=6144,
        description="Dimensionality of the 'intermediate' (often named feed-forward) layer in the Transformer encoder.",
    )

    hidden_act: str = schema_utils.StringOptions(
        options=["gelu", "relu", "silu", "tanh", "gelu_fast", "mish", "linear", "sigmoid", "gelu_new"],
        default="gelu",
        description="The non-linear activation function (function or string) in the encoder and pooler.",
    )

    hidden_dropout_prob: float = schema_utils.NonNegativeFloat(
        default=0.1,
        description="The dropout probability for all fully connected layers in the embeddings, encoder, and pooler.",
    )

    attention_probs_dropout_prob: float = schema_utils.NonNegativeFloat(
        default=0.1,
        description="The dropout ratio for the attention probabilities.",
    )

    max_position_embeddings: int = schema_utils.PositiveInteger(
        default=512,
        description=(
            "The maximum sequence length that this model might ever be used with. Typically set this to something "
            "large just in case (e.g., 512 or 1024 or 2048)."
        ),
    )

    type_vocab_size: int = schema_utils.NonNegativeInteger(
        default=0,
        description=("The vocabulary size of the `token_type_ids`."),
    )

    initializer_range: float = schema_utils.NonNegativeFloat(
        default=0.02,
        description=(
            "The standard deviation of the truncated_normal_initializer for initializing all weight matrices."
        ),
    )

    layer_norm_eps: float = schema_utils.NonNegativeFloat(
        default=1e-7,
        description="The epsilon used by the layer normalization layers.",
    )

    relative_attention: bool = schema_utils.Boolean(
        default=True,
        description="Whether use relative position encoding.",
    )

    max_relative_positions: int = schema_utils.Integer(
        default=-1,
        description=(
            "The range of relative positions `[-max_position_embeddings, max_position_embeddings]`. Use the same "
            "value as `max_position_embeddings`."
        ),
    )

    pad_token_id: int = schema_utils.Integer(
        default=0,
        description="The value used to pad input_ids.",
    )

    position_biased_input: bool = schema_utils.Boolean(
        default=False,
        description="Whether add absolute position embedding to content embedding.",
    )

    pos_att_type: List[str] = schema_utils.List(
        default=["p2c", "c2p"],
        description=(
            "The type of relative position attention, it can be a combination of `['p2c', 'c2p']`, e.g. `['p2c']`, "
            "`['p2c', 'c2p']`, `['p2c', 'c2p']`."
        ),
    )

    layer_norm_eps: float = schema_utils.NonNegativeFloat(
        default=1e-12,
        description="The epsilon used by the layer normalization layers.",
    )

    pooler_hidden_size: int = schema_utils.PositiveInteger(
        default=1536,
        description="The hidden size of the pooler layers.",
    )

    pooler_dropout: float = schema_utils.NonNegativeFloat(
        default=0,
        description="The dropout ratio for the pooler layers.",
    )

    pooler_hidden_act: str = schema_utils.StringOptions(
        options=["gelu", "relu", "silu", "tanh", "gelu_fast", "mish", "linear", "sigmoid", "gelu_new"],
        default="gelu",
        description="The activation function (function or string) in the pooler.",
    )

    position_buckets: int = schema_utils.PositiveInteger(
        default=256,
        description="The number of buckets to use for each attention layer.",
    )

    share_att_key: bool = schema_utils.Boolean(
        default=True,
        description="Whether to share attention key across layers.",
    )

    norm_rel_ebd: str = schema_utils.StringOptions(
        options=["layer_norm", "none"],
        default="layer_norm",
        description="The normalization method for relative embeddings.",
    )



from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.registry import DEFAULT_KEYS, Registry

_trainers_registry = Registry()
_ray_trainers_registry = Registry()

_llm_trainers_registry = Registry()
_llm_ray_trainers_registry = Registry()


@DeveloperAPI
def get_trainers_registry() -> Registry:
    return _trainers_registry


@DeveloperAPI
def get_ray_trainers_registry() -> Registry:
    return _ray_trainers_registry


@DeveloperAPI
def get_llm_trainers_registry() -> Registry:
    return _llm_trainers_registry


@DeveloperAPI
def get_llm_ray_trainers_registry() -> Registry:
    return _llm_ray_trainers_registry


@DeveloperAPI
def register_trainer(model_type: str, default=False):
    """Register a trainer class that supports training the given model types.

    Using default=True will make the trainer the default trainer for the model type.

    Args:
        model_type: The model_type which dictates the trainer type to use.
        default: Whether the trainer should be the default trainer for the model type.
    """

    def wrap(cls):
        _trainers_registry[model_type] = cls
        if default:
            if DEFAULT_KEYS[0] in _trainers_registry:
                raise ValueError(f"Default trainer already registered for model type {model_type}")
            for key in DEFAULT_KEYS:
                _trainers_registry[key] = cls
        return cls

    return wrap


@DeveloperAPI
def register_ray_trainer(model_type: str, default=False):
    """Register a trainer class that supports training the given model types with Ray backend.

    Using default=True will make the trainer the default trainer for the model type.

    Args:
        model_type: The model_type which dictates the trainer type to use.
        default: Whether the trainer should be the default trainer for the model type.
    """

    def wrap(cls):
        _ray_trainers_registry[model_type] = cls
        if default:
            if DEFAULT_KEYS[0] in _ray_trainers_registry:
                raise ValueError(f"Default trainer already registered for model type {model_type}")
            for key in DEFAULT_KEYS:
                _ray_trainers_registry[key] = cls
        return cls

    return wrap


@DeveloperAPI
def register_llm_trainer(trainer_type: str, default=False):
    """Register a trainer class that supports training the specific type of training strategy for LLM Models.

    Using default=True will make the trainer the default trainer for the LLM model type.

    Args:
        trainer_type: The trainer_type which dictates what training strategy to use.
        default: Whether the trainer should be the default trainer for LLMs.
    """

    def wrap(cls):
        _llm_trainers_registry[trainer_type] = cls
        if default:
            if DEFAULT_KEYS[0] in _trainers_registry:
                raise ValueError(f"Default trainer {trainer_type} already registered for LLM")
            for key in DEFAULT_KEYS:
                _llm_trainers_registry[key] = cls
        return cls

    return wrap


@DeveloperAPI
def register_llm_ray_trainer(trainer_type: str, default=False):
    """Register a trainer class that supports training the specific type of training strategy for LLM Models with
    Ray backend.

    Using default=True will make the trainer the default trainer for the LLM model type.

    Args:
        trainer_type: The trainer_type which dictates what training strategy to use.
        default: Whether the trainer should be the default trainer for LLMs.
    """

    def wrap(cls):
        _llm_ray_trainers_registry[trainer_type] = cls
        if default:
            if DEFAULT_KEYS[0] in _trainers_registry:
                raise ValueError(f"Default ray trainer {trainer_type} already registered for LLM")
            for key in DEFAULT_KEYS:
                _llm_ray_trainers_registry[key] = cls
        return cls

    return wrap



# register all decoders
import ludwig.decoders.generic_decoders  # noqa
import ludwig.decoders.llm_decoders  # noqa
import ludwig.decoders.sequence_decoders  # noqa
import ludwig.decoders.sequence_tagger  # noqa



import pytest
import torch
from transformers import AutoTokenizer

from ludwig.constants import LOGITS, PREDICTIONS, PROBABILITIES
from ludwig.utils.llm_utils import (
    add_left_padding,
    create_attention_mask,
    find_last_matching_index,
    generate_merged_ids,
    has_padding_token,
    pad_target_tensor_for_fine_tuning,
    realign_target_and_prediction_tensors_for_inference,
    remove_left_padding,
    set_pad_token,
)

pytestmark = [pytest.mark.llm]

# Pad token ID is 1 for OPT even though it uses the GPT2 tokenizer
# BOS token ID is 2
TEST_MODEL_NAME = "hf-internal-testing/tiny-random-OPTForCausalLM"


@pytest.fixture
def tokenizer():
    tokenizer = AutoTokenizer.from_pretrained(TEST_MODEL_NAME)
    set_pad_token(tokenizer)
    return tokenizer


@pytest.fixture
def input_ids():
    # Provide sample input IDs tensor
    return torch.tensor([[3, 4, 5], [6, 7, 8]])


@pytest.fixture
def target_ids():
    # Provide sample target IDs tensor
    return torch.tensor([[9, 10, 11], [12, 13, 14]])


def test_set_pad_token_doesnt_exist():
    tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)
    assert tokenizer.pad_token_id is None

    set_pad_token(tokenizer)
    assert tokenizer.pad_token_id == 50256


def test_set_pad_token_already_exists():
    tokenizer = AutoTokenizer.from_pretrained(TEST_MODEL_NAME, use_fast=False)
    assert tokenizer.pad_token_id == 1

    set_pad_token(tokenizer)
    assert tokenizer.pad_token_id == 1


def test_has_padding_token_with_padding_tokens(tokenizer):
    input_sentence = "This is an example sentence."
    input_ids = tokenizer([input_sentence])
    input_ids["input_ids"] = torch.tensor(input_ids["input_ids"])
    padded_input_ids = torch.nn.functional.pad(input_ids["input_ids"], (10 - len(input_ids["input_ids"]), 1), value=1)

    assert has_padding_token(padded_input_ids, tokenizer)


def test_has_padding_token_without_padding_tokens(tokenizer):
    input_sentence = "This is an example sentence."
    input_ids = tokenizer([input_sentence])
    input_ids["input_ids"] = torch.tensor(input_ids["input_ids"])

    assert not has_padding_token(input_ids["input_ids"], tokenizer)


@pytest.mark.parametrize(
    "input_ids, expected",
    [
        # No padding
        (torch.tensor([5]), torch.tensor([5])),
        (torch.tensor([5, 3]), torch.tensor([5, 3])),
        # Padding
        (torch.tensor([1, 5, 5, 3]), torch.tensor([5, 5, 3])),
        # EOS token
        (torch.tensor([2, 5, 5, 3]), torch.tensor([2, 5, 5, 3])),
        # Padding + EOS token
        (torch.tensor([1, 2, 5, 5, 3]), torch.tensor([2, 5, 5, 3])),
    ],
)
def test_remove_left_padding(input_ids, expected, tokenizer):
    assert torch.equal(remove_left_padding(input_ids, tokenizer).squeeze(0), expected)


@pytest.mark.parametrize(
    "input_ids, max_length, pad_value, expected",
    [
        (torch.tensor([1, 2, 3]), 3, 0, torch.tensor([1, 2, 3])),
        (torch.tensor([1, 2, 3]), 5, 0, torch.tensor([0, 0, 1, 2, 3])),
        (torch.tensor([4, 5, 6, 7]), 6, 2, torch.tensor([2, 2, 4, 5, 6, 7])),
        (torch.tensor([8, 9]), 3, 1, torch.tensor([1, 8, 9])),
    ],
)
def test_add_left_padding(input_ids, max_length, pad_value, expected):
    padded = add_left_padding(input_ids, max_length, pad_value).squeeze(0)

    assert torch.equal(padded, expected)


def test_create_attention_mask_last_token_padding(tokenizer):
    input_ids = torch.tensor([3, 4, tokenizer.pad_token_id])
    attention_mask = create_attention_mask(input_ids, tokenizer)
    assert attention_mask[-1] == 1


@pytest.mark.parametrize(
    "input_ids, expected_output",
    [
        # No padding
        (torch.tensor([3, 4, 5]), torch.tensor([1, 1, 1])),
        (torch.tensor([1, 1, 4, 6, 8]), torch.tensor([0, 0, 1, 1, 1])),
        # All padding
        (torch.tensor([1, 1, 1]), torch.tensor([0, 0, 1])),
    ],
)
def test_create_attention_mask(input_ids, expected_output, tokenizer):
    attention_mask = create_attention_mask(input_ids, tokenizer)

    assert torch.equal(attention_mask, expected_output)


@pytest.mark.parametrize(
    "tensor_a, tensor_b, expected_index",
    [
        # Matching index at the end
        (torch.tensor([1, 2, 3, 4, 5, 6, 7, 8]), torch.tensor([6, 7, 8]), 5),
        # No matching index
        (torch.tensor([1, 2, 3, 4, 5, 6, 7, 8]), torch.tensor([9, 10]), -1),
        # Matching index in the middle. Fails because we're only checking the last X elements.
        (torch.tensor([1, 2, 3, 4, 5, 6, 7, 8]), torch.tensor([4, 5, 6]), -1),
    ],
)
def test_find_last_matching_index(tensor_a, tensor_b, expected_index):
    last_matching_index = find_last_matching_index(tensor_a, tensor_b)
    assert last_matching_index == expected_index


def test_generate_merged_ids_with_target(tokenizer, input_ids, target_ids):
    # Test case when target_ids is not None
    merged_ids, attention_masks = generate_merged_ids(input_ids, target_ids, tokenizer)
    assert torch.equal(merged_ids, torch.tensor([[3, 4, 5, 9, 10, 11, 1], [6, 7, 8, 12, 13, 14, 1]]))
    assert merged_ids.shape == (2, 7)  # Check the shape of merged_ids
    assert attention_masks.shape == (2, 7)  # Check the shape of attention_masks


def test_generate_merged_ids_with_max_sequence_length(tokenizer, input_ids, target_ids):
    # Test case when max_sequence_length is provided
    max_sequence_length = 5
    merged_ids, attention_masks = generate_merged_ids(input_ids, target_ids, tokenizer, max_sequence_length)
    assert merged_ids.shape == (2, 5)  # Check the shape of merged_ids with truncation
    assert attention_masks.shape == (2, 5)  # Check the shape of attention_masks


def test_generate_merged_ids_padding_removal(tokenizer, input_ids, target_ids):
    # Test case to check removal of left padding from inputs and targets during merge
    padding_tokens = torch.tensor([tokenizer.pad_token_id, tokenizer.pad_token_id])

    # Adds 2 padding tokens to the left of input_ids and target_ids individually. Typically, if we just merged this
    # naively, we would expect to see [1, 1, 3, 4, 5, 1, 1, 9, 10, 11, 1], but we shouldn't see the padding tokens
    # except for the padding token at the end.
    input_ids_with_padding = torch.cat((padding_tokens.unsqueeze(0).expand(input_ids.size(0), -1), input_ids), dim=1)
    target_ids_with_padding = torch.cat((padding_tokens.unsqueeze(0).expand(target_ids.size(0), -1), target_ids), dim=1)

    merged_ids, attention_masks = generate_merged_ids(input_ids_with_padding, target_ids_with_padding, tokenizer)

    assert merged_ids.shape == (2, 7)  # Check the shape of merged_ids
    assert attention_masks.shape == (2, 7)  # Check the shape of attention_masks

    assert torch.equal(merged_ids[0][:3], input_ids[0])  # Check the input_ids part without padding
    assert torch.equal(merged_ids[0][3:-1], target_ids[0])  # Check the target_ids part without padding
    assert torch.equal(merged_ids[0][-1], torch.tensor(tokenizer.pad_token_id))  # Check the padding tokens

    assert torch.all(attention_masks == 1)


def test_generate_merged_ids_returns_tensor(tokenizer, input_ids, target_ids):
    # Test that the function returns torch.Tensor objects
    merged_ids, attention_masks = generate_merged_ids(input_ids, target_ids, tokenizer)
    assert isinstance(merged_ids, torch.Tensor)
    assert isinstance(attention_masks, torch.Tensor)


def test_pad_target_tensor_for_fine_tuning():
    of_name = "out_1"
    prediction = {
        of_name: {PREDICTIONS: torch.tensor([[764, 764, 764, 764, 764, 764, 764, 578, 619, 841, 182, 905, 483, 764]])}
    }

    # Scenario 1: Entire target tensor was passed into model inputs
    model_input = torch.tensor([[0, 0, 24, 52, 654, 529, 221, 78, 79, 504, 76, 397, 84, 0]])
    target = {of_name: torch.tensor([[78, 79, 504, 76, 397, 84, 0]])}
    expected_target = {of_name: torch.tensor([[-100, -100, -100, -100, -100, -100, -100, 78, 79, 504, 76, 397, 84, 0]])}
    updated_targets = pad_target_tensor_for_fine_tuning(target, prediction, model_input, of_name)
    assert torch.equal(expected_target[of_name], updated_targets[of_name])

    # Scenario 2: Entire target tensor was not passed into model inputs
    model_input = torch.tensor([[13, 24, 395, 13, 46, 57, 52, 41, 45, 37, 51, 14, 380, 435]])
    target = {of_name: torch.tensor([[78, 79, 504, 76, 397, 84, 0]])}
    expected_target = {
        of_name: torch.tensor([[-100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100]])
    }
    updated_targets = pad_target_tensor_for_fine_tuning(target, prediction, model_input, of_name)
    assert torch.equal(expected_target[of_name], updated_targets[of_name])

    # Scenario 3: Partial target tensor was passed into model inputs
    model_input = torch.tensor([[0, 0, 24, 52, 654, 529, 221, 78, 79, 504, 76, 78, 79, 504]])
    target = {of_name: torch.tensor([[78, 79, 504, 76, 397, 84, 0]])}
    expected_target = {
        of_name: torch.tensor([[-100, -100, -100, -100, -100, -100, -100, -100, -100, -100, -100, 78, 79, 504]])
    }
    updated_targets = pad_target_tensor_for_fine_tuning(target, prediction, model_input, of_name)
    assert torch.equal(expected_target[of_name], updated_targets[of_name])


def test_realign_target_and_prediction_tensors_for_inference(tokenizer):
    of_name = "out_1"
    vocab_size = 8

    # Scenario 1: Prediction and target tensors have the same length, so nothing should change
    targets = {of_name: torch.tensor([[78, 79, 504, 76, 397, 84, 0]])}
    predictions = {
        of_name: {
            PREDICTIONS: torch.tensor([[78, 79, 504, 76, 397, 84, 0]], dtype=torch.int64),
            PROBABILITIES: torch.randn(1, 7, vocab_size).to(torch.float32),
            LOGITS: torch.randn(1, 7, vocab_size).to(torch.float32),
        }
    }
    updated_targets, updated_predictions = realign_target_and_prediction_tensors_for_inference(
        targets, predictions, of_name, tokenizer
    )

    assert targets == updated_targets
    assert predictions == updated_predictions
    assert predictions[of_name][PREDICTIONS].shape[1] == targets[of_name].shape[1]
    assert predictions[of_name][PROBABILITIES].shape[1] == targets[of_name].shape[1]
    assert predictions[of_name][LOGITS].shape[1] == targets[of_name].shape[1]

    # Scenario 2: Prediction length is longer than the target tensor, so we need to realign the target tensor
    targets = {of_name: torch.tensor([[78, 79, 504, 76, 397, 84, 0]])}
    predictions = {
        of_name: {
            PREDICTIONS: torch.tensor([[98, 47, 78, 79, 504, 76, 397, 84, 0]], dtype=torch.int64),
            PROBABILITIES: torch.randn(1, 9, vocab_size).to(torch.float32),
            LOGITS: torch.randn(1, 9, vocab_size).to(torch.float32),
        }
    }
    updated_targets, updated_predictions = realign_target_and_prediction_tensors_for_inference(
        targets, predictions, of_name, tokenizer
    )

    assert predictions == updated_predictions
    assert torch.equal(updated_targets[of_name], torch.tensor([[78, 79, 504, 76, 397, 84, 0, 1, 1]]))

    # Scenario 3: Target length is longer than the prediction tensor, so we need to realign them
    targets = {of_name: torch.tensor([[98, 47, 78, 79, 504, 76, 397, 84, 0]])}
    predictions = {
        of_name: {
            PREDICTIONS: torch.tensor([[78, 79, 504, 76, 397, 84, 0]], dtype=torch.int64),
            PROBABILITIES: torch.randn(1, 7, vocab_size).to(torch.float32),
            LOGITS: torch.randn(1, 7, vocab_size).to(torch.float32),
        }
    }
    updated_targets, updated_predictions = realign_target_and_prediction_tensors_for_inference(
        targets, predictions, of_name, tokenizer
    )

    assert targets == updated_targets

    assert torch.equal(updated_predictions[of_name][PREDICTIONS], torch.tensor([[78, 79, 504, 76, 397, 84, 0, 1, 1]]))
    assert updated_predictions[of_name][PROBABILITIES].shape[1] == targets[of_name].shape[1]
    assert updated_predictions[of_name][LOGITS].shape[1] == targets[of_name].shape[1]

    assert torch.equal(updated_predictions[of_name][PROBABILITIES][0][-1], torch.zeros(vocab_size))
    assert torch.equal(updated_predictions[of_name][PROBABILITIES][0][-2], torch.zeros(vocab_size))
    assert not torch.equal(updated_predictions[of_name][PROBABILITIES][0][-3], torch.zeros(vocab_size))

    assert torch.equal(updated_predictions[of_name][LOGITS][0][-1], torch.zeros(vocab_size))
    assert torch.equal(updated_predictions[of_name][LOGITS][0][-2], torch.zeros(vocab_size))
    assert not torch.equal(updated_predictions[of_name][LOGITS][0][-3], torch.zeros(vocab_size))



import copy
import logging
import sys
import warnings
from typing import Any, Dict, List, Mapping, Set, TYPE_CHECKING

from marshmallow import ValidationError

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    CATEGORY,
    COMBINED,
    DECODER,
    DEFAULTS,
    ENCODER,
    GRID_SEARCH,
    INPUT_FEATURES,
    LOSS,
    MODEL_ECD,
    OUTPUT_FEATURES,
    PARAMETERS,
    PREPROCESSING,
    SEQUENCE,
    SPACE,
    TEXT,
    TYPE,
)
from ludwig.features.feature_utils import compute_feature_hash
from ludwig.schema.features.utils import output_config_registry
from ludwig.schema.hyperopt.scheduler import BaseHyperbandSchedulerConfig
from ludwig.schema.trainer import ECDTrainerConfig
from ludwig.types import HyperoptConfigDict, ModelConfigDict
from ludwig.utils.data_utils import get_sanitized_feature_name

if TYPE_CHECKING:
    from ludwig.schema.model_types.base import ModelConfig


logger = logging.getLogger(__name__)


@DeveloperAPI
def merge_with_defaults(config_dict: ModelConfigDict) -> ModelConfigDict:
    # Recursive merge of the features, except that if we find a dictionary containing
    # an explicit "type" key, we ignore defaults if they don't match.
    defaults = config_dict.get(DEFAULTS)
    if not defaults:
        return config_dict

    config_dict = copy.deepcopy(config_dict)
    _merge_features_(config_dict.get(INPUT_FEATURES, []), defaults, {DECODER, LOSS})
    _merge_features_(config_dict.get(OUTPUT_FEATURES, []), defaults, {ENCODER, PREPROCESSING})
    return config_dict


def _merge_features_(features: List[Dict[str, Any]], defaults: Dict[str, Any], exclude_keys: Set[str]):
    for feature in features:
        ftype = feature.get(TYPE)
        if not ftype:
            continue

        default_feature = defaults.get(ftype, {})
        merged_feature = _merge_dict_with_types(default_feature, feature, exclude_keys)

        # In-place replacement of the old feature with the new
        feature.clear()
        feature.update(merged_feature)


def _merge_dict_with_types(dct: Dict[str, Any], merge_dct: Dict[str, Any], exclude_keys: Set[str]) -> Dict[str, Any]:
    dct = copy.deepcopy(dct)
    dct = {k: v for k, v in dct.items() if k not in exclude_keys}

    for k, v in merge_dct.items():
        # TODO(travis): below type comparison is not perfect, as it doesn't consider the case where the default type
        # is omitted while the encoder type is explicitly set to the default type, in which case they
        # should resolve to equal, but will be considered different.
        if (
            k in dct
            and isinstance(dct[k], dict)
            and isinstance(v, Mapping)
            and dct[k].get(TYPE) == v.get(TYPE, dct[k].get(TYPE))
        ):
            dct[k] = _merge_dict_with_types(dct[k], v, exclude_keys)
        else:
            dct[k] = v
    return dct


@DeveloperAPI
def merge_fixed_preprocessing_params(config: "ModelConfig"):
    """Update preprocessing parameters if encoders require fixed preprocessing parameters."""
    for feature in config.input_features:
        feature.encoder.set_fixed_preprocessing_params(config.model_type, feature.preprocessing)


def set_validation_parameters(config: "ModelConfig"):
    """Sets validation-related parameters used for early stopping, determining the best hyperopt trial, etc."""
    if not config.output_features:
        return

    # First set the validation field so we know what feature we're validating on
    if not config.trainer.validation_field:
        if config.trainer.validation_metric is None or config.trainer.validation_metric == LOSS:
            # Loss is valid for all features.
            config.trainer.validation_field = config.output_features[0].name
        else:
            # Determine the proper validation field for the user, like if the user specifies "accuracy" but forgets to
            # change the validation field from "combined" to the name of the feature that produces accuracy metrics.
            from ludwig.utils.metric_utils import get_feature_to_metric_names_map

            feature_to_metric_names_map = get_feature_to_metric_names_map(config.output_features.to_list())
            validation_field = None
            for feature_name, metric_names in feature_to_metric_names_map.items():
                if config.trainer.validation_metric in metric_names:
                    if validation_field is None:
                        validation_field = feature_name
                    else:
                        raise ValidationError(
                            f"The validation_metric: '{config.trainer.validation_metric}' corresponds to multiple "
                            f"possible validation_fields, '{validation_field}' and '{feature_name}'. Please explicitly "
                            "specify the validation_field that should be used with the validation_metric "
                            f"'{config.trainer.validation_metric}'."
                        )
            if validation_field is None:
                raise ValidationError("User-specified trainer.validation_metric is not valid for any output feature.")

            config.trainer.validation_field = validation_field

    # If the field is combined, then make sure the metric is loss and then return
    if config.trainer.validation_field == COMBINED:
        # Only loss is supported for combined
        if not config.trainer.validation_metric:
            config.trainer.validation_metric = LOSS
        elif config.trainer.validation_metric != LOSS:
            raise ValidationError(
                f"Must set validation_metric=loss when validation_field=combined, "
                f"found validation_metric={config.trainer.validation_metric}"
            )
        return

    # Field is not combined, so use the default validation metric for the single feature
    validation_features = [f for f in config.output_features if f.name == config.trainer.validation_field]
    if len(validation_features) > 1:
        raise ValidationError(
            f"Found more than one feature matching validation field: {config.trainer.validation_field}"
        )
    if len(validation_features) == 0:
        raise ValidationError(f"No output feature found matching validation field: {config.trainer.validation_field}")

    validation_feature = validation_features[0]
    if not config.trainer.validation_metric:
        # The user has not explicitly set any validation fields.
        # Default to using the first output feature's default validation metric.
        out_type = validation_feature.type
        config.trainer.validation_metric = output_config_registry(config.model_type)[out_type].default_validation_metric


def set_derived_feature_columns_(config_obj: "ModelConfig"):
    """Assigns column and proc_column values to features that do not have them set.

    Proc_column is set to a hash of the feature's preprocessing configuration.
    """
    for feature in config_obj.input_features:
        if feature.column is None:
            feature.column = feature.name
        if feature.proc_column is None:
            feature.proc_column = compute_feature_hash(feature.to_dict())

    for feature in config_obj.output_features:
        if feature.column is None:
            feature.column = feature.name
        if feature.proc_column is None:
            feature.proc_column = compute_feature_hash(feature.to_dict())


def sanitize_and_filter_combiner_entities_(config: "ModelConfig"):
    if config.model_type != MODEL_ECD or config.combiner.type != "comparator":
        return

    input_feature_names = {input_feature.name for input_feature in config.input_features}

    # Sanitize feature names.
    config.combiner.entity_1 = [get_sanitized_feature_name(fname) for fname in config.combiner.entity_1]
    config.combiner.entity_2 = [get_sanitized_feature_name(fname) for fname in config.combiner.entity_2]

    entity_1_excluded = {fname for fname in config.combiner.entity_1 if fname not in input_feature_names}
    if entity_1_excluded:
        logger.warning(
            f"Excluding `entity_1` features {entity_1_excluded} from the comparator combiner because they are not "
            f"present in the `input_features`."
        )

    config.combiner.entity_1 = [fname for fname in config.combiner.entity_1 if fname not in entity_1_excluded]

    entity_2_excluded = {fname for fname in config.combiner.entity_2 if fname not in input_feature_names}
    if entity_2_excluded:
        logger.warning(
            f"Excluding `entity_2` features {entity_2_excluded} from the comparator combiner because they are not "
            f"present in the `input_features`."
        )

    config.combiner.entity_2 = [fname for fname in config.combiner.entity_2 if fname not in entity_2_excluded]


def set_hyperopt_defaults_(config: "ModelConfig"):
    """This function was migrated from defaults.py with the intention of setting some hyperopt defaults while the
    hyperopt section of the config object is not fully complete.

    Returns:
        None -> modifies trainer and hyperopt sections
    """
    if not config.hyperopt:
        return

    # Set default num_samples based on search space if not set by user
    if config.hyperopt.executor.num_samples is None:
        _contains_grid_search_params = contains_grid_search_parameters(config.hyperopt.to_dict())
        if _contains_grid_search_params:
            logger.info(
                "Setting hyperopt num_samples to 1 to prevent duplicate trials from being run. Duplicate trials are"
                " created when there are hyperopt parameters that use the `grid_search` search space.",
            )
            config.hyperopt.executor.num_samples = 1
        else:
            logger.info("Setting hyperopt num_samples to 10.")
            config.hyperopt.executor.num_samples = 10

    scheduler = config.hyperopt.executor.scheduler
    if scheduler.type == "fifo":
        # FIFO scheduler has no constraints
        return

    # Disable early stopping when using a scheduler. We achieve this by setting the parameter
    # to -1, which ensures the condition to apply early stopping is never met.
    early_stop = config.trainer.early_stop
    if early_stop is not None and early_stop != -1:
        warnings.warn("Can't utilize `early_stop` while using a hyperopt scheduler. Setting early stop to -1.")
    config.trainer.early_stop = -1

    if isinstance(config.trainer, ECDTrainerConfig) and isinstance(scheduler, BaseHyperbandSchedulerConfig):
        # TODO(travis): explore similar contraints for GBMs, which don't have epochs
        max_t = scheduler.max_t
        time_attr = scheduler.time_attr
        epochs = config.trainer.epochs
        if max_t is not None:
            if time_attr == "time_total_s":
                if epochs is None:
                    # Continue training until time limit hit
                    config.trainer.epochs = sys.maxsize
                # else continue training until either time or trainer epochs limit hit
            elif epochs is not None and epochs != max_t:
                raise ValueError(
                    "Cannot set trainer `epochs` when using hyperopt scheduler w/different training_iteration `max_t`. "
                    "Unset one of these parameters in your config or make sure their values match."
                )
            else:
                # Run trainer until scheduler epochs limit hit
                config.trainer.epochs = max_t
        elif epochs is not None:
            scheduler.max_t = epochs  # run scheduler until trainer epochs limit hit


def set_preprocessing_parameters(config: "ModelConfig") -> None:  # noqa: F821
    """Reconcile conflicting preprocessing parameters in place."""
    _set_max_sequence_length(config)


def _set_max_sequence_length(config: "ModelConfig") -> None:  # noqa: F821
    """Ensures that `max_sequence_length` is never less than `sequence_length`."""

    types_with_sequence_length = [SEQUENCE, TEXT]
    for input_feature in config.input_features:
        if input_feature.type in types_with_sequence_length:
            sequence_length = input_feature.preprocessing.sequence_length
            max_sequence_length = input_feature.preprocessing.max_sequence_length
            if sequence_length is not None and sequence_length > max_sequence_length:
                warnings.warn(
                    "if `sequence_length` is not None, `max_sequence_length` must be greater than or equal "
                    "to `sequence_length`. Setting `max_sequence_length` to `sequence_length`."
                )
                input_feature.preprocessing.max_sequence_length = sequence_length


def set_tagger_decoder_parameters(config: "ModelConfig") -> None:
    """Overrides the reduce_input parameter for text and sequence output features when a tagger decoder is used.
    This is done to ensure that the decoder correctly gets a 3D tensor as input.

    Returns:
        None -> modifies output_features
    """
    for output_feature in config.output_features:
        if output_feature.type in {TEXT, SEQUENCE} and output_feature.decoder.type == "tagger":
            if output_feature.reduce_input is not None:
                warnings.warn(
                    "reduce_input must be set to `None` when using a tagger decoder for your output feature. "
                    f"Setting reduce_input to `None` for `{output_feature.name}`."
                )
                output_feature.reduce_input = None


def set_llm_tokenizers(config: "ModelConfig") -> None:
    """Sets the tokenizers for the LLM model to the pretrained model name or path. This ensures that they use the
    correct shared vocabulary from the tokenizer.

    This also ensures padding is correctly set to left padding to prevent the LLM from trying to continue to sequence
    based on the right padding tokens, which might exist based on sequence length.
    """
    if config.model_type != "llm":
        return

    pretrained_model_name_or_path = config.base_model
    if not isinstance(pretrained_model_name_or_path, str) or pretrained_model_name_or_path is None:
        raise ValueError("Must set `base_model` when using the LLM model.")

    for input_feature in config.input_features:
        if input_feature.type == TEXT:
            input_feature.preprocessing.tokenizer = "hf_tokenizer"
            input_feature.preprocessing.pretrained_model_name_or_path = pretrained_model_name_or_path
            input_feature.preprocessing.padding = "left"

    for output_feature in config.output_features:
        if output_feature.type == TEXT:
            # Add tokenizer parameters to preprocessing so it can be used during post processing
            output_feature.preprocessing.tokenizer = "hf_tokenizer"
            output_feature.preprocessing.pretrained_model_name_or_path = pretrained_model_name_or_path
            output_feature.preprocessing.padding = "left"

            # Add tokenizer parameters to decoder so it can be used during the forward pass
            output_feature.decoder.pretrained_model_name_or_path = pretrained_model_name_or_path
            output_feature.decoder.max_new_tokens = config.generation.max_new_tokens
        elif output_feature.type == CATEGORY:
            # Tokenizer parameters
            output_feature.decoder.tokenizer = "hf_tokenizer"
            output_feature.decoder.pretrained_model_name_or_path = pretrained_model_name_or_path
            # Parameters for building decoder vocabulary
            output_feature.decoder.fallback_label = output_feature.preprocessing.fallback_label


@DeveloperAPI
def contains_grid_search_parameters(hyperopt_config: HyperoptConfigDict) -> bool:
    """Returns True if any hyperopt parameter in the config is using the grid_search space."""
    for _, param_info in hyperopt_config[PARAMETERS].items():
        if param_info.get(SPACE, None) == GRID_SEARCH:
            return True
    return False



import hashlib
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Type, TYPE_CHECKING, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from ludwig.vector_index import FAISS, get_vector_index_cls
from ludwig.vector_index.base import VectorIndex

if TYPE_CHECKING:
    from sentence_transformers import SentenceTransformer
    from ludwig.backend.base import Backend

from ludwig.utils.batch_size_tuner import BatchSizeEvaluator
from ludwig.utils.torch_utils import get_torch_device


def df_checksum(df: pd.DataFrame) -> str:
    return hashlib.sha1(pd.util.hash_pandas_object(df).values).hexdigest()


def df_to_row_strs(df: pd.DataFrame) -> List[str]:
    rows = df.to_dict(orient="records")
    row_strs = [json.dumps(r) for r in rows]
    return row_strs


class RetrievalModel(ABC):
    @abstractmethod
    def create_dataset_index(self, df: pd.DataFrame, backend: "Backend", columns_to_index: Optional[List[str]] = None):
        """Creates an index for the dataset.

        If `columns_to_index` is None, all columns are indexed. Otherwise, only the columns in `columns_to_index` are
        used for indexing, but all columns in `df` are returned in the search results.
        """
        pass

    @abstractmethod
    def search(
        self, df, backend: "Backend", k: int = 10, return_data: bool = False
    ) -> Union[List[int], List[Dict[str, Any]]]:
        """Retrieve the top k results for the given query.

        If `return_data` is True, returns the data associated with the indices. Otherwise, returns the indices.
        """
        pass

    @abstractmethod
    def save_index(self, name: str, cache_directory: str):
        """Saves the index to the cache directory."""
        pass

    @abstractmethod
    def load_index(self, name: str, cache_directory: str):
        """Loads the index from the cache directory."""
        pass


class RandomRetrieval(RetrievalModel):
    """Random retrieval model.

    Gets k random indices from the dataset regardless of the query.
    """

    def __init__(self, **kwargs):
        self.index = None
        self.index_data = None

    def create_dataset_index(self, df: pd.DataFrame, backend: "Backend", columns_to_index: Optional[List[str]] = None):
        self.index = np.array(range(len(df)))
        self.index_data = df

    def search(
        self, df, backend: "Backend", k: int = 10, return_data: bool = False
    ) -> Union[List[int], List[Dict[str, Any]]]:
        results = []
        for _ in tqdm(range(len(df))):
            indices = np.random.choice(self.index, k, replace=False)

            if return_data:
                result = self.index_data.iloc[indices].to_dict(orient="records")
            else:
                result = indices
            results.append(result)
        return results

    def save_index(self, name: str, cache_directory: str):
        index_file_path = os.path.join(cache_directory, name + ".index")
        # open file to prevent using the .npy extension
        # https://numpy.org/doc/stable/reference/generated/numpy.save.html
        with open(index_file_path, "wb") as f:
            np.save(f, self.index)

        index_data_file_path = os.path.join(cache_directory, name + "_data.csv")
        self.index_data.to_csv(index_data_file_path, index=False)

    def load_index(self, name: str, cache_directory: str):
        index_file_path = os.path.join(cache_directory, name + ".index")
        self.index = np.load(index_file_path)

        index_data_file_path = os.path.join(cache_directory, name + "_data.csv")
        self.index_data = pd.read_csv(index_data_file_path)


class SemanticRetrieval(RetrievalModel):
    """Semantic retrieval model.

    Uses a sentence transformer model to encode the dataset and retrieve the top k most similar results to the query.
    """

    def __init__(self, model_name, **kwargs):
        self.model_name = model_name
        self.model = get_semantic_retrieval_model(self.model_name)
        self.index: VectorIndex = None
        self.index_data: pd.DataFrame = None

        # best batch size computed during the encoding step
        self.best_batch_size = None

    def create_dataset_index(self, df: pd.DataFrame, backend: "Backend", columns_to_index: Optional[List[str]] = None):
        if columns_to_index is None:
            columns_to_index = df.columns
        df_to_index = df[columns_to_index]
        row_strs = df_to_row_strs(df_to_index)

        embeddings = self._encode(row_strs, backend)
        self.index = get_vector_index_cls(FAISS).from_embeddings(embeddings)
        # Save the entire df so we can return the full row when searching
        self.index_data = df

    def _encode(self, row_strs: List[str], backend: "Backend") -> np.ndarray:
        # only do this step once
        if self.best_batch_size is None:
            self.best_batch_size = backend.tune_batch_size(
                create_semantic_retrieval_model_evaluator(self.model, row_strs), len(row_strs)
            )

        transform_fn = create_semantic_retrieval_model_fn(self.model, self.best_batch_size)
        df = backend.df_engine.from_pandas(pd.DataFrame({"data": row_strs}))
        df = backend.batch_transform(df, self.best_batch_size, transform_fn)
        df = backend.df_engine.compute(df)
        embeddings = np.stack(df["data"].values).astype(np.float32)
        return embeddings

    def search(
        self, df: pd.DataFrame, backend: "Backend", k: int = 10, return_data: bool = False
    ) -> Union[List[int], List[Dict[str, Any]]]:
        row_strs = df_to_row_strs(df)

        query_vectors = self._encode(row_strs, backend)
        results = []
        # TODO(geoffrey): figure out why self.index.search segfaults with larger batch sizes
        for query_vector in tqdm(query_vectors, total=query_vectors.shape[0]):
            indices = self.index.search(query_vector.reshape(1, -1), k)
            if return_data:
                result = self.index_data.iloc[indices].to_dict(orient="records")
            else:
                result = indices
            results.append(result)
        return results

    def save_index(self, name: str, cache_directory: str):
        index_file_path = os.path.join(cache_directory, name + ".index")
        self.index.save(index_file_path)

        index_data_file_path = os.path.join(cache_directory, name + "_data.csv")
        self.index_data.to_csv(index_data_file_path, index=False)

    def load_index(self, name: str, cache_directory: str):
        index_file_path = os.path.join(cache_directory, name + ".index")
        self.index = get_vector_index_cls(FAISS).from_path(index_file_path)

        index_data_file_path = os.path.join(cache_directory, name + "_data.csv")
        self.index_data = pd.read_csv(index_data_file_path)


def create_semantic_retrieval_model_evaluator(
    model: "SentenceTransformer", samples: List[str]
) -> Type[BatchSizeEvaluator]:
    class _RetrievalModelEvaluator(BatchSizeEvaluator):
        def __init__(self):
            self.model = model.to(get_torch_device())
            self.samples = samples

        def step(self, batch_size: int):
            self.model.encode(self.samples[:batch_size], batch_size=batch_size, show_progress_bar=False)

    return _RetrievalModelEvaluator


def create_semantic_retrieval_model_fn(
    model: "SentenceTransformer", batch_size: int
) -> Callable[[pd.DataFrame], np.ndarray]:
    class _RetrievalModelFn:
        def __init__(self):
            self.model = model.to(get_torch_device())
            self.batch_size = batch_size

        def __call__(self, df: pd.DataFrame) -> np.ndarray:
            row_strs = df["data"].tolist()
            result = self.model.encode(row_strs, batch_size=self.batch_size, show_progress_bar=False)
            df["data"] = result.tolist()
            return df

    return _RetrievalModelFn


def get_semantic_retrieval_model(model_name: str) -> "SentenceTransformer":
    from sentence_transformers import SentenceTransformer

    return SentenceTransformer(model_name, device=get_torch_device())


def get_retrieval_model(type: str, **kwargs) -> RetrievalModel:
    if type == "random":
        return RandomRetrieval(**kwargs)
    elif type == "semantic":
        return SemanticRetrieval(**kwargs)
    else:
        raise ValueError(f"Unsupported retrieval model type: {type}")



import ludwig.schema.combiners.comparator  # noqa: F401
import ludwig.schema.combiners.concat  # noqa: F401
import ludwig.schema.combiners.project_aggregate  # noqa: F401
import ludwig.schema.combiners.sequence  # noqa: F401
import ludwig.schema.combiners.sequence_concat  # noqa: F401
import ludwig.schema.combiners.tab_transformer  # noqa: F401
import ludwig.schema.combiners.tabnet  # noqa: F401
import ludwig.schema.combiners.transformer  # noqa: F401
